"""
BamTools - Utilities for working with BAM files
===============================================

This module brings together convenience function for working
with :term:`bam` formatted files.

"""

from pysam.libchtslib cimport *
from pysam.libcalignmentfile cimport *
from pysam.libcalignedsegment cimport pysam_bam_get_cigar, \
    pysam_bam_get_qname, pysam_get_n_cigar
from pysam.libcfaidx cimport *
from libc.string cimport strchr
from libc.stdint cimport int8_t
from libc.stdio cimport puts, printf
from libc.stdlib cimport abs
from cpython cimport PyErr_SetString, PyBytes_FromStringAndSize
from cpython cimport array as c_array
from sortedcontainers import SortedList

import array
import base64
import collections
import copy
import struct
import hashlib
import itertools
import numpy
import pandas
import pysam
import sys

cimport numpy

import cgatcore.experiment as E

def parse_region_string(s):
    """parse a genomic region string.

    Returns tuple of contig, start, end. Missing values are None.
    """
    if not s:
        return None, None, None
    if ":" in s:
        contig, coords = s.split(":")
        if "-" in coords:
            start, end = list(map(int, coords.split("-")))
        else:
            start = int(coords)
            end = None
        return contig, start, end
    else:
        return s, None, None


FLAGS = {
    1: 'paired',
    2: 'proper_pair',
    4: 'unmapped',
    8: 'mate_unmapped',
    16: 'reverse',
    32: 'mate_reverse',
    64: 'read1',
    128: 'read2',
    256: 'secondary',
    512: 'qc_fail',
    1024: 'duplicate',
    2048: 'supplementary'}

cdef struct CountsType:
    int32_t read_length
    int8_t alignments
    int8_t is_mapped
    int8_t is_unmapped
    int8_t mate_is_unmapped
    int8_t is_paired
    int8_t mapped_is_read1
    int8_t mapped_is_read2
    int8_t unmapped_is_read1
    int8_t unmapped_is_read2
    int8_t is_proper_pair
    int8_t is_secondary
    int8_t is_qcfail
    int8_t is_duplicate
    int8_t is_supplementary
    int32_t mean_quality
    int32_t median_quality

cdef int NM = 10

# pass views and fill instead of creating.
# This saves time, but also prevents possible memory leaks.
cdef int NCIGAR_CODES = 10
cdef get_cigar_stats(AlignedSegment read, uint32_t[:] base_counts, uint32_t[:] block_counts):
    cdef int nfields = NCIGAR_CODES + 1

    cdef bam1_t * src = read._delegate
    cdef int op
    cdef uint32_t l
    cdef int32_t k
    cdef uint32_t * cigar_p = pysam_bam_get_cigar(src)

    for k from 0 <= k < nfields:
        base_counts[k] = 0
        block_counts[k] = 0

    if cigar_p == NULL:
        return None

    for k from 0 <= k < pysam_get_n_cigar(src):
        op = cigar_p[k] & BAM_CIGAR_MASK
        l = cigar_p[k] >> BAM_CIGAR_SHIFT
        base_counts[op] += l
        block_counts[op] += 1

    cdef uint8_t * v = bam_aux_get(src, 'NM')
    if v != NULL:
        base_counts[nfields - 1] = <int32_t>bam_aux2i(v)

    return base_counts, block_counts


def bam2stats_count(AlignmentFile samfile,
                    bed_mask=None,
                    ignore_masked_reads=False,
                    is_stdin=True,
                    filename_fastq=None,
                    outfile_details=None,
                    add_alignment_details=False,
                    outfile_readmap=None,
                    detailed_count=None):
    '''
    '''
    cdef AlignedSegment read
    cdef bint _add_alignment_details = add_alignment_details
    # counters
    cdef int ninput = 0
    cdef int nduplicates = 0
    # number of reads present after filtering
    cdef int nfiltered = 0
    # number of reads overlapping masked regions (if bed_mask != None)
    cdef int nmasked = 0
    # number of reads not overlap RNA (if bed_mask != None)
    cdef int nnotmasked = 0

    cdef bint _ignore_masked_reads = ignore_masked_reads
    cdef int max_hi = 0
    # count nh, nm tags
    nh_filtered = collections.defaultdict(int)
    nm_filtered = collections.defaultdict(int)
    nh_all = collections.defaultdict(int)
    nm_all = collections.defaultdict(int)
    mapq_filtered = collections.defaultdict(int)
    mapq_all = collections.defaultdict(int)

    cdef int * flags_counts = <int*>calloc(len(FLAGS), sizeof(int))

    # helper variables
    cdef int last_tid = -1
    cdef int last_pos = 0
    cdef int32_t read_index

    duplicates = collections.defaultdict(int)
    counts = collections.defaultdict(int)
    cdef int count = 0

    cdef int tid = -1
    cdef int flag
    cdef uint32_t iteration = 0
    cdef uint32_t report_step = 1000000
    cdef uint32_t nalignments = 0
    cdef uint8_t * v
    cdef int32_t nm
    cdef int32_t nh
    cdef int32_t hi
    cdef int x
    cdef int lflags = len(FLAGS)
    cdef uint32_t f

    # Todo: make this filter configurable
    cdef int detail_filter_flags = 2304

    # detailed counting
    cdef FastxRecord fq
    cdef int64_t index, fastq_nreads
    cdef CountsType * fastq_counts
    cdef CountsType * fastq_count
    cdef char * read_name
    cdef char * position
    cdef bint count_fastq = False
    cdef int fastq_notfound = 0
    cdef int chop = 0
    cdef int hash_size = 5

    # number of alignment pairs that are nucleotide mismatches
    cdef long mismatch_counts = 0
    # number of alignment pairs where there is an insertion in the reference
    cdef long insertion_counts = 0
    # number of alignment pairs where there is a deletion in the reference
    cdef long deletion_counts = 0
    # number of alignment pairs that are in match state
    # (might be a nucleotide mismatch)
    cdef long match_counts = 0

    cdef uint16_t [:, :] alignment_details_view

    cdef int nfields = NCIGAR_CODES + 1

    cdef c_array.array base_counts = array.array(
        "I",
         [0] * nfields)
    cdef c_array.array block_counts = array.array(
             "I",
        [0] * nfields)

    cdef uint32_t [:] base_counts_view = base_counts
    cdef uint32_t [:] block_counts_view = block_counts

    if filename_fastq != None:
        count_fastq = True
        E.info("reading fastq file")
        # Using a python dictionary here is due
        # for a large amount of memory usage.
        # Alternatives to dictionary
        # 1. POSIX hash tables (hsearch,...) or trees: very slow
        # 2. custom hash implementation: worth the effort?
        # 3. Sorted list and binary search: too slow for many lookups
        reads = {}
        fastqfile = FastxFile(filename_fastq)
        fastq_nreads = 0
        for fq in fastqfile:
            if fastq_nreads == 0:
                # chop off /1 or /2 as mappers usually remove these
                # suffices. Test only the first.
                name = fq.name
                if name.endswith("/1") or name.endswith("/2"):
                    chop = -2
                
            if chop != 0:
                name = hashlib.md5(fq.name[:chop].encode("ascii")).digest()[:hash_size]
            else:
                name = hashlib.md5(fq.name.encode("ascii")).digest()[:hash_size]

            reads[name] = fastq_nreads

            fastq_nreads += 1
        
    elif not is_stdin and detailed_count:
        count_fastq = True
        old_pos = samfile.tell()
        E.info("two-pass processing to collect read names")
        reads = {}
        fastq_nreads = 0
        for iteration, read in enumerate(samfile):

            if iteration % report_step == 0:
                E.info("pre-pass: read {} alignments: {} reads".format(
                    iteration, len(reads)))

            read_name = pysam_bam_get_qname(read._delegate)
            
            # terminate string at first space to
            # truncate read names containing more than
            # just the id
            position = strchr(read_name, ' ')
            if position != NULL: 
                position[0] = '\0'

            # compute hash, truncate to half the md5 digest which
            # should be 64 bits and provides 36,893,488,147,419,103,232
            # hashes.
            md5 = hashlib.md5(read_name).digest()[:hash_size]
            if md5 not in reads:
                reads[md5] = fastq_nreads
                fastq_nreads += 1
                if outfile_readmap:
                    outfile_readmap.write("{}\t{}\n".format(
                        read_name,
                        base64.encodestring(md5)[:-1]))

        samfile.seek(old_pos)
        nalignments = iteration
    else:
        count_fastq = False
        E.info("simple counting only")

    if count_fastq:
        E.info("read names of %i reads or read pairs" % fastq_nreads)
        E.info("allocating %i bytes" % (fastq_nreads * sizeof(CountsType)))

        fastq_counts = <CountsType *>calloc(fastq_nreads, sizeof(CountsType))
        if fastq_counts == NULL:
            raise ValueError("could not allocate memory for %i bytes" %
                             (fastq_nreads * sizeof(CountsType)))

    if _add_alignment_details:
        alignment_details = numpy.zeros((fastq_nreads, NCIGAR_CODES + 1), dtype=numpy.uint16)
                
    E.info("starting processing of alignment file")

    for iteration, read in enumerate(samfile):

        if iteration % report_step == 0:
            if nalignments:
                E.info("read {}/{} alignments: {}%".format(
                    iteration, nalignments,
                    100.0 * iteration / nalignments))
            else:
                E.info("read {} alignments".format(
                    iteration))

        flag = read._delegate.core.flag
        ninput += 1

        f = 1
        for x from 0 <= x < lflags:
            if flag & f:
                flags_counts[x] += 1
            f = f << 1

        # get maximum NI field
        v = bam_aux_get(read._delegate, 'HI')
        if v != NULL:
            hi = <int32_t>bam_aux2i(v)
            if hi > max_hi: max_hi = hi

        v = bam_aux_get(read._delegate, 'NH')
        if v != NULL:
            nh = <int32_t>bam_aux2i(v)
            nh_all[nh] += 1
        else:
            nh = -1

        v = bam_aux_get(read._delegate, 'NM')
        if v != NULL:
            nm = <int32_t>bam_aux2i(v)
            nm_all[nm] += 1
        else:
            nm = -1

        mapq_all[read.mapq] += 1

        get_cigar_stats(read, base_counts_view, block_counts_view) 

        if not read.is_secondary:
            # for consistency with samtools
            mismatch_counts += nm
            deletion_counts += base_counts_view[BAM_CDEL]
            insertion_counts += base_counts_view[BAM_CINS]
            match_counts += base_counts_view[BAM_CMATCH]

        if count_fastq:
            read_name = pysam_bam_get_qname(read._delegate)
            # terminate string at first space to
            # truncate read names containing more than
            # just the id
            position = strchr(read_name, ' ')
            if position != NULL: 
                position[0] = '\0'

            md5 = hashlib.md5(read_name).digest()[:hash_size]
            read_index = reads[md5]
            try:
                fastq_count = &fastq_counts[read_index]
            except KeyError:
                fastq_notfound += 1
                continue

            # only take primary alignments for read length. read.query_length
            # includes soft-clipped sequence.
            if not read.is_secondary and not read.is_supplementary:
                fastq_count.read_length = read.query_length
                q = read.query_qualities
                # multiply by 1000 to increase significant digits
                if q:
                    fastq_count.mean_quality = round(numpy.mean(q) * 1000.0)
                    fastq_count.median_quality = round(numpy.median(q) * 1000.0)
                else:
                    fastq_count.mean_quality = 0
                    fastq_count.median_quality = 0

            # book-keeping counts
            fastq_count.alignments += 1

            if read.is_qcfail: fastq_count.is_qcfail += 1
            if read.is_duplicate: fastq_count.is_duplicate += 1
            if read.is_paired: fastq_count.is_paired += 1
                
            if read.is_unmapped: 
                fastq_count.is_unmapped += 1
                if read.is_read1: fastq_count.unmapped_is_read1 += 1
                if read.is_read2: fastq_count.unmapped_is_read2 += 1
            else:
                # only count is_read1 and is_read2 for mapped
                fastq_count.is_mapped += 1
                if read.mate_is_unmapped: fastq_count.mate_is_unmapped += 1
                if read.is_read1: fastq_count.mapped_is_read1 += 1
                if read.is_read2: fastq_count.mapped_is_read2 += 1
                if read.is_proper_pair: fastq_count.is_proper_pair += 1
                if read.is_secondary:
                    fastq_count.is_secondary += 1
                if read.is_supplementary:
                    fastq_count.is_supplementary += 1
                if _add_alignment_details:
                    if not (flag & detail_filter_flags):
                        alignment_details[read_index, :] += base_counts_view

        # is paired and read2
        #if flag & 1 and flag & 128:
            # ignore second reads to avoid double counting pairs
            # this needs to be done properly by counting in
            # pairs
        #    continue

        # skip unmapped reads
        if read._delegate.core.flag & 4:
            continue

        if read.tid != last_tid:
            contig = samfile.getrname(read.rname)

        # note: does not take into account gaps within reads
        # or partial overlap.
        if bed_mask:
            if bed_mask.contains(contig, read.pos, read.pos + read.alen):
                nmasked += 1
                if _ignore_masked_reads:
                    continue
            else:
                nnotmasked += 1
            
        nfiltered += 1

        if nh >= 0: nh_filtered[nh] += 1
        if nm >= 0: nm_filtered[nm] += 1
        mapq_filtered[read.mapq] += 1

        # duplicate analysis - simply count per start position
        # ignoring sequence and strand
        if read.tid == last_tid and read.pos == last_pos:
            count += 1
            nduplicates += 1
            continue

        if count > 1:
            counts[count] += 1

        count = 1
        last_tid, last_pos = read.tid, read.pos

    E.info( "finished computing counts" )

    if fastq_notfound:
        E.warn( "could not match %i records in bam-file to fastq file" % fastq_notfound)

    counter = E.Counter()

    counter.alignments_input = ninput
    counter.alignments_filtered = nfiltered
    counter.alignments_duplicates = nduplicates
    counter.alignments_masked = nmasked
    counter.alignments_notmasked = nnotmasked
    
    if match_counts == 0:
        match_counts = 1

    counter.error_counts = mismatch_counts + insertion_counts
    counter.match_counts = match_counts
    counter.mismatch_counts = mismatch_counts
    counter.deletion_counts = deletion_counts
    counter.insertion_counts = insertion_counts

    counter.error_rate = float(mismatch_counts) / (match_counts + insertion_counts)
    counter.match_rate = float(match_counts) / (match_counts + insertion_counts)
    counter.mismatch_rate = float(mismatch_counts) / (match_counts)
    counter.deletion_rate = float(deletion_counts) / (match_counts + deletion_counts)
    counter.insertion_rate = float(insertion_counts) / (match_counts + insertion_counts)

    # convert flags to labels
    t = {}
    f = 1
    for x from 0 <= x < lflags:
        t[FLAGS[f]] = flags_counts[x]
        f = f << 1

    # count based on fastq data
    cdef int total_paired = 0
    cdef int total_unpaired = 0
    cdef int total_pair_is_mapped = 0
    cdef int total_pair_is_unmapped = 0
    cdef int total_pair_is_proper_uniq = 0
    cdef int total_pair_is_proper_mmap = 0
    cdef int total_pair_is_proper_duplicate = 0
    cdef int total_pair_not_proper_uniq = 0
    cdef int total_pair_is_incomplete_uniq = 0
    cdef int total_pair_is_incomplete_mmap = 0
    cdef int total_pair_is_other = 0
    # read based counting for unpaired reads
    cdef int total_reads = 0
    cdef int total_read_is_mapped_uniquely = 0
    cdef int total_read_is_mapped_multiply = 0
    cdef int total_read_is_unmapped = 0
    cdef int total_read_is_missing = 0
    cdef int total_read_has_supplementary = 0
    # read based counting for read1
    cdef int total_read1 = 0
    cdef int total_read1_is_mapped_uniquely = 0
    cdef int total_read1_is_mapped_multiply = 0
    cdef int total_read1_is_unmapped = 0
    cdef int total_read1_is_missing = 0
    # read based counting for read2
    cdef int total_read2 = 0
    cdef int total_read2_is_mapped_uniquely = 0
    cdef int total_read2_is_mapped_multiply = 0
    cdef int total_read2_is_unmapped = 0
    cdef int total_read2_is_missing = 0

    cdef float substitution_rate
    cdef float insertion_rate
    cdef float deletion_rate
    cdef float error_rate
    cdef float coverage
        
    if count_fastq:
        E.info("fastq counting: aggregating counts")
        for qname, index in reads.items():
            fastq_count = &fastq_counts[index]

            # paired read data
            if fastq_count.is_paired: 
                total_paired += 1

                if fastq_count.is_unmapped == fastq_count.is_paired:
                    # an unmapped read pair
                    total_pair_is_unmapped += 1
                    continue

                total_pair_is_mapped += 1

                if fastq_count.is_proper_pair == 2: 
                    # a unique proper pair
                    total_pair_is_proper_uniq +=1
                    # a duplicate unique proper pair
                    if fastq_count.is_duplicate == 2:
                        total_pair_is_proper_duplicate +=1
                elif fastq_count.is_proper_pair > 2:
                    # proper pairs that map to multiple locations
                    total_pair_is_proper_mmap +=1
                elif fastq_count.mapped_is_read1 == 1 \
                        and fastq_count.mapped_is_read2 == 1 \
                        and fastq_count.is_proper_pair == 0:
                    # pair which map each read once, but not is not proper
                    total_pair_not_proper_uniq += 1
                elif (fastq_count.mapped_is_read1 == 1 and
                      fastq_count.mapped_is_read2 == 0) \
                    or (fastq_count.mapped_is_read1 == 0 and
                        fastq_count.mapped_is_read2 == 1):
                    # an incomplete pair - one read of a pair matches uniquely 
                    # but not the other
                    total_pair_is_incomplete_uniq += 1
                elif (fastq_count.mapped_is_read1 == 1 and
                      fastq_count.mapped_is_read2 > 1) \
                    or (fastq_count.mapped_is_read1 > 1 and
                        fastq_count.mapped_is_read2 == 1):
                    # an incomplete pair - one read of a pair matches uniquely
                    # but the other matches multiple times
                    total_pair_is_incomplete_mmap += 1
                else:
                    total_pair_is_other += 1

                # paired read counting
                if fastq_count.mapped_is_read1: 
                    if fastq_count.mapped_is_read1 == 1:
                        total_read1_is_mapped_uniquely += 1
                    elif fastq_count.mapped_is_read1 > 1:
                        total_read1_is_mapped_multiply += 1
                elif fastq_count.unmapped_is_read1 > 0:
                    total_read1_is_unmapped += 1
                else:
                    total_read1_is_missing += 1

                if fastq_count.mapped_is_read2: 
                    if fastq_count.mapped_is_read2 == 1:
                        total_read2_is_mapped_uniquely += 1
                    elif fastq_count.mapped_is_read2 > 1:
                        total_read2_is_mapped_multiply += 1
                elif fastq_count.unmapped_is_read2 > 0:
                    total_read2_is_unmapped += 1
                else:
                    total_read2_is_missing += 1
                        
            else:
                # reads without data, could be unpaired or missing
                total_unpaired += 1

                if fastq_count.is_supplementary > 0:
                    total_read_has_supplementary += 1

                # counting for unpaired data
                if fastq_count.is_unmapped > 0:
                    total_read_is_unmapped += 1
                else:
                    if fastq_count.is_mapped == 1:
                        total_read_is_mapped_uniquely += 1
                    elif fastq_count.is_mapped > 1:
                        total_read_is_mapped_multiply += 1
                    else:
                        total_read_is_missing += 1
                    
        ###################################################
        # Pair statistics
        counter.total_pairs = total_paired + total_unpaired
        counter.total_pair_is_mapped = total_pair_is_mapped
        counter.total_pair_is_unmapped = total_pair_is_unmapped
        counter.total_pair_is_proper_uniq = total_pair_is_proper_uniq
        counter.total_pair_is_incomplete_uniq = total_pair_is_incomplete_uniq
        counter.total_pair_is_incomplete_mmap = total_pair_is_incomplete_mmap
        counter.total_pair_is_proper_duplicate = total_pair_is_proper_duplicate
        counter.total_pair_is_proper_mmap = total_pair_is_proper_mmap
        counter.total_pair_not_proper_uniq = total_pair_not_proper_uniq
        counter.total_pair_is_other = total_pair_is_other

        ###################################################
        # Read Stats for SE and PE data
        # for SE data
        if total_paired == 0:
            E.info("reads: single end counting")
            counter.total_read = fastq_nreads
            counter.total_read_is_mapped_uniq = total_read_is_mapped_uniquely
            counter.total_read_is_mmap = total_read_is_mapped_multiply
            counter.total_read_is_mapped = total_read_is_mapped_multiply +\
                                           total_read_is_mapped_uniquely
            counter.total_read_is_unmapped = total_read_is_unmapped
            counter.total_read_is_missing = total_read_is_missing
            counter.total_read_has_supplementary = total_read_has_supplementary
        else:
            # for PE data
            E.info("reads: paired end counting: paired = %i" % total_paired)
            counter.total_read = fastq_nreads * 2
            counter.total_read_is_mapped_uniq = total_read1_is_mapped_uniquely +\
                                                total_read2_is_mapped_uniquely
            counter.total_read_is_mmap = total_read1_is_mapped_multiply +\
                                         total_read2_is_mapped_multiply
            counter.total_read_is_mapped = counter.total_read_is_mapped_uniq +\
                                           counter.total_read_is_mmap
            counter.total_read_is_unmapped = total_read1_is_unmapped +\
                                             total_read2_is_unmapped
            counter.total_read_is_missing = total_read1_is_missing +\
                                            total_read2_is_missing
            counter.total_read_has_supplementary = total_read_has_supplementary * 2

        ###################################################
        ## stats for 1st/2nd read separately
        counter.total_read1 = fastq_nreads
        counter.total_read1_is_mapped_uniq = total_read1_is_mapped_uniquely
        counter.total_read1_is_mmap = total_read1_is_mapped_multiply
        counter.total_read1_is_mapped = counter.total_read1_is_mapped_uniq + counter.total_read1_is_mmap
        counter.total_read1_is_unmapped = total_read1_is_unmapped
        counter.total_read1_is_missing = total_read1_is_missing
        counter.total_read2 = fastq_nreads
        counter.total_read2_is_mapped_uniq = total_read2_is_mapped_uniquely
        counter.total_read2_is_mmap = total_read2_is_mapped_multiply
        counter.total_read2_is_mapped = counter.total_read2_is_mapped_uniq + counter.total_read2_is_mmap
        counter.total_read2_is_unmapped = total_read2_is_unmapped
        counter.total_read2_is_missing = total_read2_is_missing

        substitution_rates = numpy.zeros(fastq_nreads, dtype=numpy.float)
        insertion_rates = numpy.zeros(fastq_nreads, dtype=numpy.float)
        deletion_rates = numpy.zeros(fastq_nreads, dtype=numpy.float)
        error_rates = numpy.zeros(fastq_nreads, dtype=numpy.float)
        coverages = numpy.zeros(fastq_nreads, dtype=numpy.float)
        mask = numpy.ones(fastq_nreads, dtype=numpy.int)

        if outfile_details:
            header = ["read_md5",
                      "read_length",
                      "alignments",
                      "mean_quality",
                      "median_quality",
                      "is_unmapped",
                      "mate_is_unmapped",
                      "is_paired",
                      "mapped_is_read1",
                      "mapped_is_read2",
                      "is_proper_pair",
                      "is_secondary",
                      "is_qcfail", 
                      "is_duplicate",
                      "is_supplementary"]

            if _add_alignment_details:
                header.extend([
                    "cigar_match", "cigar_ins",
                    "cigar_del", "cigar_ref_skip",
                    "cigar_soft_clip", "cigar_hard_clip",
                    "cigar_pad", "cigar_equal",
                    "cigar_diff", "cigar_back",
                    "mismatches",
                    "coverage",
                    "substitution_rate",
                    "insertion_rate",
                    "deletion_rate",
                    "error_rate"])

            if outfile_details != sys.stdout:
                # later: get access FILE * object
                outfile_details.write("\t".join(header) + "\n")
                for qname, read_index in reads.items():
                    fastq_count = &fastq_counts[read_index]
                    # remove "\n" from base64 encoded md5
                    outfile_details.write("%s\t%s" % (
                        base64.encodestring(qname)[:-1].decode("ascii"),
                        "\t".join( \
                                   map(str,
                                       (fastq_count.read_length,
                                        fastq_count.alignments,
                                        "{:5.2f}".format(fastq_count.mean_quality / 1000.0),
                                        "{:5.2f}".format(fastq_count.median_quality / 1000.0),
                                        fastq_count.is_unmapped,
                                        fastq_count.mate_is_unmapped,
                                        fastq_count.is_paired,
                                        fastq_count.mapped_is_read1,
                                        fastq_count.mapped_is_read2,
                                        fastq_count.is_proper_pair,
                                        fastq_count.is_secondary,
                                        fastq_count.is_qcfail,
                                        fastq_count.is_duplicate,
                                        fastq_count.is_supplementary
                                    ) )) ))
                    if _add_alignment_details:
                        if fastq_count.is_unmapped > 0:
                            substitution_rate = 0
                            insertion_rate = 0
                            deletion_rate = 0
                            error_rate = 0
                            coverage = 0
                            mask[read_index] = 0
                        else:
                            substitution_rate = float(alignment_details[read_index, NM]) /\
                                (alignment_details[read_index, BAM_CMATCH] +
                                 alignment_details[read_index, BAM_CINS])
                            insertion_rate = float(alignment_details[read_index, BAM_CINS]) /\
                                (alignment_details[read_index, BAM_CMATCH] +
                                 alignment_details[read_index, BAM_CINS])
                            deletion_rate = float(alignment_details[read_index, BAM_CDEL]) /\
                                            (alignment_details[read_index, BAM_CMATCH] +
                                             alignment_details[read_index, BAM_CDEL])
                            error_rate = float(alignment_details[read_index, NM] + \
                                          alignment_details[read_index, BAM_CINS]) /\
                                (alignment_details[read_index, BAM_CMATCH] +
                                 alignment_details[read_index, BAM_CINS])

                            # TODO: take into account supplementary aligments and soft/hard-masking?
                            if fastq_count.read_length > 0:
                                coverage = float(alignment_details[read_index, BAM_CMATCH]) /\
                                           fastq_count.read_length
                            else:
                                coverage = float(alignment_details[read_index, BAM_CMATCH]) /\
                                            (alignment_details[read_index, BAM_CMATCH] +
                                             alignment_details[read_index, BAM_CDEL])
                               
                            # ignore reads > 32k as counters might have wrapped over
                            if fastq_count.read_length >= 32767:
                                mask[read_index] = 0

                            substitution_rates[read_index] = substitution_rate
                            insertion_rates[read_index] = insertion_rate
                            deletion_rates[read_index] = deletion_rate
                            error_rates[read_index] = error_rate
                            coverages[read_index] = coverage

                        outfile_details.write(
                            "\t" +
                            "\t".join(map(str, alignment_details[read_index, :])) +
                            "\t{:6.4f}".format(coverage) +
                            "\t{:6.4f}".format(substitution_rate) +
                            "\t{:6.4f}".format(insertion_rate) +
                            "\t{:6.4f}".format(deletion_rate) +
                            "\t{:6.4f}".format(error_rate) +
                            "\n")
                    else:
                        outfile_details.write("\n")
            else:
                # TODO: this code is incomplete.
                # output to stdout much quicker
                # use puts to avoid the following error:
                # format not a string literal and no format arguments
                puts("\t".join(header) + "\n")
                for qname, read_index in reads.items():
                    fastq_count = &fastq_counts[read_index]
                    read_name = qname
                    printf("%s\t%i\t%i\t%5.2f\t%5.2f\t%i\t%i\t%i\t%i\t%i\t%i\t%i\t%i\t%i\t%i",
                           read_name,
                           fastq_count.read_length,
                           fastq_count.alignments,
                           fastq_count.mean_quality / 1000.0,
                           fastq_count.median_quality / 1000.0,
                           fastq_count.is_unmapped,
                           fastq_count.mate_is_unmapped,
                           fastq_count.is_paired,
                           fastq_count.mapped_is_read1,
                           fastq_count.mapped_is_read2,
                           fastq_count.is_proper_pair,
                           fastq_count.is_secondary,
                           fastq_count.is_qcfail,
                           fastq_count.is_duplicate,
                           fastq_count.is_supplementary)
                    if _add_alignment_details:
                        printf("\t%i\t%i\t%i\t%i\t%i\t%i\t%i\t%i\t%i\t%i\n",
                               alignment_details_view[read_index, 0],
                               alignment_details_view[read_index, 1],
                               alignment_details_view[read_index, 2],
                               alignment_details_view[read_index, 3],
                               alignment_details_view[read_index, 4],
                               alignment_details_view[read_index, 5],
                               alignment_details_view[read_index, 6],
                               alignment_details_view[read_index, 7],
                               alignment_details_view[read_index, 8],
                               alignment_details_view[read_index, 9],
                               alignment_details_view[read_index, 10])

                    else:
                        printf("\n")

        details_df = pandas.DataFrame.from_items(
            zip(["substitution_rate",
                 "insertion_rate",
                 "deletion_rate",
                 "error_rate",
                 "coverage"],
                [substitution_rates,
                 insertion_rates,
                 deletion_rates,
                 error_rates,
                 coverages]))
        # subset to only take mapped reads
        details_df = details_df[mask == 1]
    else:
        details_df = None

    return (counter, t, 
            nh_filtered,
            nh_all, 
            nm_filtered,
            nm_all, 
            mapq_filtered,
            mapq_all, 
            max_hi,
            details_df)


class BufferedBAMOutput(object):
    """receive a list of reads and output reads in 
    sorted order."""
    
    def __init__(self, samfile):
        self.samfile = samfile
        self.buffer = SortedList(key = lambda x: x.reference_start)

    def write(self, readlist):
        cdef AlignedSegment read
        for read in readlist:
            self.samfile.write(read)

    def add(self, read):
        self.buffer.add(read)

    def flush(self, read):
        
        cdef uint32_t pos
        if len(self.buffer) == 0:
            return

        # chromosome change
        if read.reference_id != self.buffer[0].reference_id:
            self.write(self.buffer)
            self.buffer = SortedList(key = lambda x: x.reference_start)
        else:
            to_output = []
            pos = read.reference_start
            while self.buffer and self.buffer[0].reference_start <= pos:
                to_output.append(self.buffer.pop(0))
            # E.debug("flushing buffer: {}, {} left".format(len(to_output),
            #                                               len(self.buffer)))
            self.write(to_output)
            
    def __dealloc__(self):
        self.write(self.buffer)

    def __len__(self):
        return len(self.buffer)


class DirectBAMOutput(object):
    """receive a list of reads and output reads as they arrive.
    """
    
    def __init__(self, samfile):
        self.samfile = samfile

    def add(self, AlignedSegment read):
        self.samfile.write(read)

    def flush(self, read):
        return

    def __len__(self):
        return 0


def bam2bam_split_reads(AlignmentFile pysam_in,
                        AlignmentFile pysam_out,
                        int max_read_length,
                        int default_quality_score,
                        output_mode="buffered"):

    if output_mode == "buffered":
        output_writer = BufferedBAMOutput(pysam_out)
    elif output_mode == "direct":
        output_writer = DirectBAMOutput(pysam_out)
    else:
        raise ValueError("unknown output mode: {}".format(output_mode))

    cdef AlignedSegment read
    cdef AlignedSegment out_read
    cdef int l, length, state
    cdef int remainder = 0
    cdef int left_bit = 0
    cdef int total_length = 0
    cdef int sequence_pos = 0
    cdef int match_length
    cdef int num_blocks
    cdef long current_reference_pos
    cdef long ninput = 0
    cdef long nskipped_no_sequence = 0
    cdef long noutput = 0

    for read in pysam_in.fetch():

        ninput += 1
        output_writer.flush(read)
        full_cigar = read.cigar
        match_length = sum([y for x, y in full_cigar if x == 0 or x == 7 or x == 8])
        num_blocks = match_length // max_read_length

        if ninput % 1000 == 0:
            E.debug("splitting read at {}:{} with {} matches into {} blocks: cache={}".format(
                    read.reference_name,
                    read.reference_start,
                    match_length,
                    num_blocks,
                    len(output_writer)))

        full_sequence = read.query_sequence
        if full_sequence is None:
            nskipped_no_sequence += 1
            continue

        full_qualities = read.query_qualities
        if full_qualities is None:
            full_qualities = [default_quality_score] * len(full_sequence)

        read.set_tag("MD", None)

        current_cigar = []
        current_sequence = []
        current_qualities = []
        current_reference_pos  = read.reference_start
        total_length = 0
        sequence_pos = 0
        for state, length in full_cigar:
            # print state, length, total_length, max_read_length
            # print current_cigar, current_sequence, current_qualities
            if state == 0 or state == 7 or state == 8:
                # test for overshoot
                if total_length + length >= max_read_length:
                    remainder = length
                    while remainder > 0:
                        # segment length to output:
                        left_bit = max_read_length - total_length

                        # add leftover segment
                        if left_bit > 0:
                            current_cigar.append((0, left_bit))
                            current_sequence.append(
                                full_sequence[sequence_pos:sequence_pos + left_bit])
                            current_qualities.extend(
                                full_qualities[sequence_pos:sequence_pos + left_bit])
                            remainder -= left_bit
                            sequence_pos += left_bit
                            current_reference_pos += left_bit

                        # output read
                        out_read = copy.copy(read)
                        out_read.cigartuples = current_cigar
                        out_read.query_sequence = "".join(current_sequence)
                        out_read.query_qualities = current_qualities
                        noutput += 1
                        output_writer.add(out_read)

                        # add remainder to read, but at most max read_length
                        l = min(remainder, max_read_length)
                        current_cigar = [(0, l)]
                        current_sequence = [
                            full_sequence[sequence_pos:sequence_pos + l]]
                        current_qualities = list(
                            full_qualities[sequence_pos:sequence_pos + l])
                        read.reference_start = current_reference_pos
                        sequence_pos += l
                        current_reference_pos += l

                        # subtract from remainder
                        total_length = l
                        remainder -= l
                    continue

                total_length += length

            if state in (0, 1, 4, 7, 8):
                current_sequence.append(
                    full_sequence[sequence_pos:sequence_pos + length])
                current_qualities.extend(
                    full_qualities[sequence_pos:sequence_pos + length])
                sequence_pos += length
            if state in (0, 2, 7, 8):
                current_reference_pos += length

            current_cigar.append((state, length))
        else:
            # emit last bit of read
            out_read = copy.copy(read)
            out_read.cigartuples = current_cigar
            out_read.query_sequence = "".join(current_sequence)
            out_read.query_qualities = current_qualities
            noutput += 1
            output_writer.add(out_read)

    E.info("splitting completed: ninput={}, noutput={}, nskipped_no_sequence={}".format(
        ninput, noutput, nskipped_no_sequence))

cdef class SetNH:

    cdef iter
    cdef stack

    def __cinit__(self, iter ):
        self.iter = itertools.groupby(iter, lambda x: x.qname)
        self.stack = []

    def __iter__(self):
        return self

    def __next__(self): 
        """python version of next().
        """
        
        while 1:
            if self.stack:
                return self.stack.pop(0)
            else:
                key, x = next(self.iter)
                self.stack = list(x)
                nh = len(self.stack)
                for read in self.stack:
                    if not read.is_unmapped:
                        # deal with paired end reads counted
                        # as multi-mapping
                        if read.is_proper_pair and nh > 1:
                            nh -= 1
                        read.set_tag("NH", nh)


def bam2bam_filter_bam(AlignmentFile input_samfile,
                       AlignmentFile output_samfile,
                       AlignmentFile reference_samfile,
                       remove_nonunique=False,
                       remove_unique=False,
                       remove_contigs=None,
                       remove_unmapped=False,
                       remove_mismatches=False,
                       filter_mismatches=None,
                       filter_error_rate=None,
                       colour_mismatches=False,
                       int minimum_read_length=0,
                       float minimum_average_base_quality=0):

    '''To conserve memory, the tid and NM flag from *transcripts_samfile*
    are packed into memory. As a consequence, this method requires
    max(NM) < 256 (2^8) and max(tid) < 16777216 (2^24)

    If *remove_nonunique* is set, only uniquely matching reads will be
    output.

    If *remove_contigs* is given, contigs that are in remove_contigs
    will be removed. Note that this will only remove the alignment,
    but not all alignments of a particuluar read and the NH flag will
    *NOT* be updated.

    If *remove_unmapped* is given, unmapped reads will be removed.

    If *remove_mismatches* is set, only reads with number of
    mismatches better than in reference_samfile will be kept.

    If *colour_mismatches* is set, the ``CM`` tag will be used to
    count differences. By default, the ``NM`` tag is used.  The tag
    that is used needs to present in both *input_samfile* and
    *reference_samfile*.

    Detecting non-unique matches:
    
    This method first checks for the NH flag - if set, a unique match
    should have at most NH=1 hits.

    If not set, the method checks for BWA flags. Currently it checks
    if X0 is set (X0=Number of best hits found by BWA). Other relevant
    flags but currently not tested:

    * X1 = Number of suboptimal hits found by BWA
    * XT = Type: Unique/Repeat/N/Mate-sw
    '''

    cdef int ninput = 0
    cdef int nunmapped = 0
    cdef int nmismatches = 0
    cdef int noutput = 0
    cdef int nerror_rate = 0
    cdef int nnonunique = 0
    cdef int nunique = 0
    cdef int nremoved_contigs = 0
    cdef int nminimum_read_length = 0
    cdef int nminimum_average_base_quality = 0

    cdef bint c_remove_nonunique = remove_nonunique
    cdef bint c_remove_unique = remove_unique
    cdef bint c_remove_mismatches = remove_mismatches
    cdef bint c_remove_unmapped = remove_unmapped

    cdef int * remove_contig_tids 
    cdef int nremove_contig_tids = 0
    cdef AlignedSegment read
    cdef AlignedSegment match

    cdef int32_t c_filter_error_rate = 0
    cdef int32_t error_rate = 0
    if filter_error_rate is not None:
        c_filter_error_rate = filter_error_rate * 10000

    # build index
    # this method will start indexing from the current file position
    # if you decide
    cdef int ret = 1
    cdef int x
    cdef bam1_t * b = <bam1_t*>calloc(1, sizeof( bam1_t))
    cdef uint64_t pos
    cdef uint8_t * v
    cdef int32_t nm
    cdef int32_t nh
    cdef int tid
    cdef int transript_tid
    cdef long val
    cdef bint skip = 0
    cdef float total_qualities

    cdef int32_t read_mismatches = 0

    # decide which tag to use
    cdef char * nm_tag = 'NM' 
    cdef char * cm_tag = 'CM'
    cdef char * tag

    if colour_mismatches:
        tag = cm_tag
    else:
        tag = nm_tag

    if c_remove_mismatches:
        E.info( "building index" )
        if not reference_samfile:
            raise ValueError("require another bam file for mismatch filtering" )

        # L = 1 byte (unsigned char)
        def _gen(): return array.array('B') 
        index = collections.defaultdict(_gen)

        while ret > 0:
            ret = bam_read1(hts_get_bgzfp(reference_samfile.htsfile),
                            b)
            if ret > 0:
                # ignore unmapped reads
                if b.core.flag & 4:
                    continue
                qname = pysam_bam_get_qname(b)
                tid = b.core.tid
                v = bam_aux_get(b, tag)
                if v != NULL:
                    nm = <int32_t>bam_aux2i(v)
                else:
                    nm = 0
                index[qname].append(nm)

        E.info( "built index for %i reads" % len(index))
        bam_destroy1(b)

    # setup list of contigs to remove:
    if remove_contigs:
        nremove_contig_tids = len(remove_contigs)
        remove_contig_tids = <int*>malloc(sizeof(int) * nremove_contig_tids)
        for x, rname in enumerate(remove_contigs):
            remove_contig_tids[x] = input_samfile.gettid(rname)
                
    E.info( "starting filtering" )

    cdef int nfields = NCIGAR_CODES + 1

    cdef c_array.array base_counts = array.array(
        "I",
         [0] * nfields)
    cdef c_array.array block_counts = array.array(
             "I",
        [0] * nfields)

    cdef uint32_t [:] base_counts_view = base_counts
    cdef uint32_t [:] block_counts_view = block_counts

    for read in input_samfile:

        ninput += 1
        # if ninput > 10000: break

        # remove unmapped reads
        if read._delegate.core.flag & 4:
            if c_remove_unmapped:
                nunmapped += 1
            else:
                output_samfile.write(read)
            continue

        if c_filter_error_rate > 0:
            v = bam_aux_get(read._delegate, 'NM')
            if v != NULL:
                nm = <int32_t>bam_aux2i(v)
            else:
                nm = 0

            get_cigar_stats(read, base_counts_view, block_counts_view) 
            error_rate = (10000 * (nm + base_counts_view[BAM_CINS])) /\
                (base_counts_view[BAM_CMATCH] + base_counts_view[BAM_CINS])
            if error_rate > c_filter_error_rate:
                nerror_rate += 1
                continue

        # remove non-unique alignments
        if c_remove_nonunique:
            # check either NH or X0 (bwa) flag
            v = bam_aux_get(read._delegate, 'NH')
            if v == NULL:
                v = bam_aux_get(read._delegate, 'X0')
            
            if v != NULL:
                nh = <int32_t>bam_aux2i(v)
                if nh > 1: 
                    nnonunique += 1
                    continue
        # remove unique alignments
        elif c_remove_unique:
            # check either NH or X0 (bwa) flag
            v = bam_aux_get(read._delegate, 'NH')
            if v == NULL:
                v = bam_aux_get(read._delegate, 'X0')
            
            if v != NULL:
                nh = <int32_t>bam_aux2i(v)
                if nh == 1: 
                    nunique += 1
                    continue

        # remove reads matched to certain contigs
        if nremove_contig_tids:
            skip = 0
            for x from 0 <= x < nremove_contig_tids:
                if remove_contig_tids[x] == read.tid:
                    skip = 1
                    break
            if skip: 
                nremoved_contigs += 1
                continue

        # remove reads in other bam file if their are better matching
        if c_remove_mismatches:
            if read.qname in index:
                # can compress index before, depends on
                # how many reads you expect test to filter
                nm = min(index[read.qname])
                read_mismatches = read.opt(tag)
                if nm > read_mismatches: 
                    nmismatches += 1
                    continue
        
        if minimum_read_length and read.query_length < minimum_read_length:
            nminimum_read_length += 1
            continue

        if minimum_average_base_quality > 0:
            if read.query_qualities:
                total_qualities = sum(read.query_qualities)
                if total_qualities / read.query_length < minimum_average_base_quality:
                    nminimum_average_base_quality += 1
                    continue
 
        noutput += 1
        output_samfile.write(read)

    c = E.Counter()
    c.input = ninput
    c.removed_unique = nunique
    c.removed_nonunique = nnonunique
    c.removed_contigs = nremoved_contigs
    c.output = noutput
    c.removed_unmapped = nunmapped
    c.removed_mismatches = nmismatches
    c.error_rate = nerror_rate
    c.minimum_read_length = nminimum_read_length
    c.minimum_average_base_quality = nminimum_average_base_quality

    if nremove_contig_tids:
        free(remove_contig_tids)

    E.info("filtering finished")

    return c


cdef inline uint32_t get_alignment_length(bam1_t * src):
    cdef int k = 0
    cdef uint32_t l = 0
    if src == NULL:
        return 0
    cdef uint32_t * cigar_p = bam_get_cigar(src)
    if cigar_p == NULL:
        return 0
    cdef int op
    cdef int n = pysam_get_n_cigar(src)
    for k from 0 <= k < n:
        op = cigar_p[k] & BAM_CIGAR_MASK
        if op == BAM_CSOFT_CLIP or op == BAM_CHARD_CLIP:
            continue
        l += cigar_p[k] >> BAM_CIGAR_SHIFT
    return l


cdef bytes count_md_tag_mismatches(AlignedSegment read):
    """return byte string with match/mismatch information.
    
    matches: =
    mismatches: *
    insertions into reference: +
    deletions in reference: -
    
    
    """

    cdef bam1_t * src = read._delegate

    if src == NULL:
        return None

    cdef uint32_t * cigar_p = pysam_bam_get_cigar(src)
    if cigar_p == NULL:
        return None

    cdef uint32_t r_idx = 0
    cdef int op
    cdef uint32_t k, i, l, x
    cdef int nmatches = 0
    cdef int s_idx = 0

    cdef uint32_t max_len = get_alignment_length(src)
    if max_len == 0:
        raise ValueError("could not determine alignment length")

    cdef char * s = <char*>calloc(max_len + 1, sizeof(char))
    if s == NULL:
        raise ValueError(
            "could not allocated sequence of length %i" % max_len)

    for k from 0 <= k < pysam_get_n_cigar(src):
        op = cigar_p[k] & BAM_CIGAR_MASK
        l = cigar_p[k] >> BAM_CIGAR_SHIFT
        if op == BAM_CMATCH or op == BAM_CEQUAL or op == BAM_CDIFF:
            for i from 0 <= i < l:
                s[s_idx] = '='
                r_idx += 1
                s_idx += 1
        elif op == BAM_CDEL:
            for i from 0 <= i < l:
                s[s_idx] = '-'
                s_idx += 1
        elif op == BAM_CREF_SKIP:
            pass
        elif op == BAM_CINS:
            for i from 0 <= i < l:
                # encode insertions into reference as lowercase
                s[s_idx] = '+'
                r_idx += 1
                s_idx += 1
        elif op == BAM_CSOFT_CLIP:
            pass
        elif op == BAM_CHARD_CLIP:
            pass # advances neither
        elif op == BAM_CPAD:
            raise NotImplementedError(
                "Padding (BAM_CPAD, 6) is currently not supported. "
                "Please implement. Sorry about that.")

    cdef uint8_t * md_tag_ptr = bam_aux_get(src, "MD")
    if md_tag_ptr == NULL:
        seq = PyBytes_FromStringAndSize(s, s_idx)
        free(s)
        return seq

    cdef char * md_tag = <char*>bam_aux2Z(md_tag_ptr)
    cdef int md_idx = 0
    s_idx = 0

    while md_tag[md_idx] != 0:
        # c is numerical
        if md_tag[md_idx] >= 48 and md_tag[md_idx] <= 57:
            nmatches *= 10
            nmatches += md_tag[md_idx] - 48
            md_idx += 1
            continue
        else:
            # save matches up to this point, skipping insertions
            for x from 0 <= x < nmatches:
                while s[s_idx] == '+':
                    s_idx += 1
                s_idx += 1
            while s[s_idx] == '+':
                s_idx += 1

            r_idx += nmatches
            nmatches = 0
            if md_tag[md_idx] == '^':
                md_idx += 1
                while md_tag[md_idx] >= 65 and md_tag[md_idx] <= 90:
                    assert s[s_idx] == '-'
                    s_idx += 1
                    md_idx += 1
            else:
                # save mismatch and change to lower case
                s[s_idx] = '*'
                s_idx += 1
                r_idx += 1
                md_idx += 1

    # save matches up to this point, skipping insertions
    for x from 0 <= x < nmatches:
        while s[s_idx] == '+':
            s_idx += 1
        s_idx += 1
    while s[s_idx] == '+':
        s_idx += 1

    seq = PyBytes_FromStringAndSize(s, s_idx)
    free(s)

    return seq


def bam2stats_window_count(AlignmentFile samfile,
                           region=None,
                           chromosomes=None,
                           window_size=1000,
                           fasta=None):
    '''
    '''
    cdef AlignedSegment read
    cdef uint32_t idx = 0
    cdef uint32_t length = 0
    cdef int nflags = len(FLAGS)
    cdef int tid = -1
    cdef int flag
    cdef uint32_t f = 0
    cdef uint32_t offset = 0
    cdef uint32_t ref_window = 0
    cdef uint32_t x = 0
    cdef uint32_t _window_size = window_size
    cdef int ninput = 0
    cdef int nunmapped = 0
    cdef char code
    cdef int32_t ref_pos
    cdef int32_t next_window_start
    cdef int32_t window_start
    cdef int32_t window_end
    cdef char * sequence
    cdef int nmatches, nmismatches, ndeletions, ninsertios 

    cdef int idx_alignment_starts = nflags
    cdef int idx_alignment_ends = idx_alignment_starts + 1
    cdef int idx_alignments = idx_alignment_ends + 1
    cdef int idx_matches = idx_alignments + 1
    cdef int idx_mismatches = idx_matches + 1
    cdef int idx_insertions = idx_mismatches + 1
    cdef int idx_deletions = idx_insertions + 1
    cdef int idx_hard_clipped = idx_deletions + 1
    cdef int idx_hard_clipped_bases = idx_hard_clipped + 1
    cdef int idx_soft_clipped = idx_hard_clipped_bases + 1
    cdef int idx_soft_clipped_bases = idx_soft_clipped + 1
    cdef int idx_gc = idx_soft_clipped_bases + 1
    cdef int idx_at = idx_gc + 1

    # build list of column names
    # add flags to cloumns
    columns = []
    f = 1
    for x from 0 <= x < nflags:
        columns.append(FLAGS[f])
        f = f << 1
    
    columns.extend([
        "alignment_starts",
        "alignment_ends",
        "alignments",
        "matched_bases",
        "mismatched_bases",
        "inserted_bases",
        "deleted_bases",
        "hard_clipped_alignments",
        "hard_clipped_bases",
        "soft_clipped_alignments",
        "soft_clipped_bases",
        "bases_gc",
        "bases_at"])

    assert idx_at == len(columns) - 1

    # build counting matrix
    contigs = samfile.references
    lengths = samfile.lengths

    if region is not None:
        itr = samfile.fetch(region)
        contig, start, end = parse_region_string(region)

        pairs = []
        for c, l in zip(contigs, lengths):
            if c == contig:
                pairs.append((c, l))
            else:
                pairs.append((c, 0))
        contigs, lengths = zip(*pairs)
    else:
        itr = samfile.fetch()
        
    cdef uint32_t total_length = sum(lengths) + _window_size * len(lengths)
    cdef uint32_t ncontigs = len(contigs)
    cdef int ncolumns = len(columns)
    cdef numpy.ndarray offsets = numpy.zeros(
        ncontigs + 1, dtype=numpy.int32)

    for idx, length in enumerate(lengths):
        offsets[idx] = offset
        if length > 0:
            offset += (length + _window_size) // _window_size

    cdef uint32_t total_windows = offset
    offsets[idx + 1] = offset

    cdef numpy.ndarray window_counts = numpy.zeros(
        (total_windows, ncolumns),
        dtype=numpy.int32)

    cdef int32_t [:, :] window_counts_view = window_counts

    E.debug("total_windows={}, total_length={}, nflags={}, ncolumns={}, ncontigs={}".format(
            total_windows, total_length, nflags, ncolumns, ncontigs))

    E.info("computing read counts in windows")

    for read in itr:

        flag = read._delegate.core.flag
        ninput += 1

        if ninput % 1000 == 0:
            E.debug("processing read number {} at {}:{}".format(
                    ninput, read.reference_name, read.reference_start))
        
        tid = read.tid
        if tid < 0:
            continue

        if flag & 4:
            nunmapped += 1
            continue

        window_start = offsets[tid] + read.reference_start // _window_size
        window_end = offsets[tid] + read.reference_end // _window_size

        window_counts[window_start][idx_alignment_starts] += 1
        window_counts[window_end][idx_alignment_ends] += 1

        # count hard/soft-clipping
        cigar = read.cigartuples
        if cigar[0][0] == 5:
            window_counts[window_start][idx_hard_clipped] += 1
            window_counts[window_start][idx_hard_clipped_bases] += cigar[0][1]
        elif cigar[0][0] == 4:
            window_counts[window_start][idx_soft_clipped] += 1
            window_counts[window_start][idx_soft_clipped_bases] += cigar[0][1]

        if cigar[-1][0] == 5:
            window_counts[window_end][idx_hard_clipped] += 1
            window_counts[window_end][idx_hard_clipped_bases] += cigar[-1][1]
        elif cigar[-1][0] == 4:
            window_counts[window_end][idx_soft_clipped] += 1
            window_counts[window_end][idx_soft_clipped_bases] += cigar[-1][1]

        for idx from window_start <= idx <= window_end:
            window_counts[idx][idx_alignments] += 1
            f = 1
            for x from 0 <= x < nflags:
                if flag & f:
                    window_counts[idx][x] += 1
                f = f << 1

        idx = window_start
        md_seq = count_md_tag_mismatches(read)
        ref_pos = read.reference_start
        nmatches = nmismatches = ndeletions = ninsertions = 0
        # border of next window
        next_window_start = offsets[tid] + ((ref_pos // _window_size + 1) * _window_size)
        for code in md_seq:

            if code == '=':
                nmatches += 1
                ref_pos += 1
            elif code == "+":
                ninsertions += 1
            elif code == "-":
                ndeletions += 1
                ref_pos += 1
            elif code == "*":
                nmismatches += 1
                ref_pos += 1
                
            if ref_pos >= next_window_start:
                window_counts[idx][idx_matches] += nmatches
                window_counts[idx][idx_mismatches] += nmismatches
                window_counts[idx][idx_deletions] += ndeletions
                window_counts[idx][idx_insertions] += ninsertions

                next_window_start += _window_size
                idx += 1

        window_counts[idx][idx_matches] += nmatches
        window_counts[idx][idx_mismatches] += nmismatches
        window_counts[idx][idx_deletions] += ndeletions
        window_counts[idx][idx_insertions] += ninsertions

    # compute G+C content
    if fasta:
        E.info("computing G+C content windows")
        idx = 0
        for contig, length in zip(contigs, lengths):
            if length == 0:
                idx += 1
                continue

            offset = offsets[idx]
            s = fasta.fetch(contig)
            sequence = s
            for x from 0 <= x < length:
                # do not increment at x == 0
                if x and x % _window_size == 0:
                    offset += 1
                    
                code = sequence[x]
                if code > 97:
                    code -= 32

                if code == 'C' or code == 'G':
                    window_counts[offset][idx_gc] += 1
                elif code == 'A' or code == 'T':
                    window_counts[offset][idx_at] += 1

            idx += 1

    # set window coordinates
    E.info("setting window coordinates")
    chromosomes, starts, ends = [], [], []
    idx = 0
    for contig, length in zip(contigs, lengths):
        if length == 0:
            idx += 1
            continue

        offset = offsets[idx]
        next_offset = offsets[idx + 1]
        nwindows = next_offset - offset
        chromosomes.extend([contig] * nwindows)
        starts.extend(numpy.arange(0, 
                                   window_size * nwindows,
                                   window_size))
        ends.extend(numpy.arange(window_size,
                                 window_size * (nwindows + 1),
                                 window_size))
        idx += 1

    E.info("building dataframe")
    window_df = pandas.DataFrame(window_counts, 
                                 columns=columns,
                                 index=[chromosomes, starts, ends])
    window_df.index.names = ["contig", "start", "end"]

    return window_df


def is_paired(bamfile, alignments=1000):
    '''check if a `bamfile` contains paired end data

    The method reads at most the first *alignments* and returns
    True if any of the alignments are paired.
    '''

    samfile = pysam.AlignmentFile(bamfile)
    n = 0
    for read in samfile:
        if read.is_paired:
            break
        n += 1
        if n == alignments:
            break

    samfile.close()

    return n != alignments


def estimateInsertSizeDistribution(bamfile,
                                   alignments=10000,
                                   n=10,
                                   method="picard",
                                   similarity_threshold=1.0,
                                   max_chunks=1000):
    '''estimate insert size from a subset of alignments in a bam file.

    Several methods are implemented.

    picard
        The method works analogous to picard by restricting the estimates
        to a core distribution. The core distribution is defined as all
        values that lie within n-times the median absolute deviation of
        the full data set.
    convergence
        The method works similar to ``picard``, but continues reading
        `alignments` until the mean and standard deviation stabilize.
        The values returned are the median mean and median standard
        deviation encountered.

    The method `convergence` is suited to RNA-seq data, as insert sizes
    fluctuate siginificantly depending on the current region
    being looked at.

    Only mapped and proper pairs are considered in the computation.

    Returns
    -------
    mean : float
       Mean of insert sizes.
    stddev : float
       Standard deviation of insert sizes.
    npairs : int
       Number of read pairs used for the estimation
    method : string
       Estimation method
    similarity_threshold : float
       Similarity threshold to apply.
    max_chunks : int
       Maximum number of chunks of size `alignments` to be used
       in the convergence method.

    '''

    assert is_paired(bamfile), \
        'can only estimate insert size from' \
        'paired bam files'

    samfile = pysam.AlignmentFile(bamfile)

    def get_core_distribution(inserts, n):
        # compute median absolute deviation
        raw_median = numpy.median(inserts)
        raw_median_dev = numpy.median(numpy.absolute(inserts - raw_median))

        # set thresholds
        threshold_min = max(0, raw_median - n * raw_median_dev)
        threshold_max = raw_median + n * raw_median_dev

        # define core distribution
        return inserts[numpy.logical_and(inserts >= threshold_min,
                                         inserts <= threshold_max)]

    if method == "picard":

        # only get first read in pair to avoid double counting
        inserts = numpy.array(
            [read.template_length for read in samfile.head(n=alignments)
             if read.is_proper_pair
             and not read.is_unmapped
             and not read.mate_is_unmapped
             and not read.is_read1
             and not read.is_duplicate
             and read.template_length > 0])
        core = get_core_distribution(inserts, n)

        return numpy.mean(core), numpy.std(core), len(inserts)

    elif method == "convergence":

        means, stds, counts = [], [], []
        last_mean = 0
        iteration = 0
        while iteration < max_chunks:

            inserts = numpy.array(
                [read.template_length for read in samfile.head(
                    n=alignments,
                    multiple_iterators=False)
                 if read.is_proper_pair
                 and not read.is_unmapped
                 and not read.mate_is_unmapped
                 and not read.is_read1
                 and not read.is_duplicate
                 and read.template_length > 0])
            core = get_core_distribution(inserts, n)
            means.append(numpy.mean(core))
            stds.append(numpy.std(core))
            counts.append(len(inserts))
            mean_core = get_core_distribution(numpy.array(means), 2)
            mm = numpy.mean(mean_core)
            if abs(mm - last_mean) < similarity_threshold:
                break
            last_mean = mm

        return numpy.median(means), numpy.median(stds), sum(counts)
    else:
        raise ValueError("unknown method '%s'" % method)


def estimateTagSize(bamfile,
                    alignments=10,
                    multiple="error"):
    '''estimate tag/read size from first alignments in file.

    Arguments
    ---------
    bamfile : string
       Filename of :term:`bam` formatted file
    alignments : int
       Number of alignments to inspect
    multiple : string
       How to deal if there are multiple tag sizes present.
       ``error`` will raise a warning, ``mean`` will return the
       mean of the read lengths found. ``uniq`` will return a
       unique list of read sizes found. ``all`` will return all
       read sizes encountered.

    Returns
    -------
    size : int
       The read size (actual, mean or list of read sizes)

    Raises
    ------
    ValueError
       If there are multiple tag sizes present and `multiple` is set to
       `error`.

    '''
    samfile = pysam.AlignmentFile(bamfile)
    sizes = [read.rlen for read in samfile.head(alignments)]
    mi, ma = min(sizes), max(sizes)

    if mi == 0 and ma == 0:
        sizes = [read.inferred_length for read in samfile.head(alignments)]
        # remove 0 sizes (unaligned reads?)
        sizes = [x for x in sizes if x > 0]
        mi, ma = min(sizes), max(sizes)

    if mi != ma:
        if multiple == "error":
            raise ValueError('multiple tag sizes in %s: %s' % (bamfile, sizes))
        elif multiple == "mean":
            mi = int(sum(sizes) / len(sizes))
        elif multiple == "uniq":
            mi = list(sorted(set(sizes)))
        elif multiple == "all":
            return sizes

    return mi


def getNumberOfAlignments(bamfile):
    '''return number of alignments in bamfile.
    '''
    samfile = pysam.AlignmentFile(bamfile)
    return samfile.mapped


def getNumReads(bamfile):
    '''count number of reads in bam file.

    This methods works through pysam.idxstats.

    Arguments
    ---------
    bamfile : string
        Filename of :term:`bam` formatted file. The file needs
        to be indexed.
    Returns
    -------
    nreads : int
        Number of reads
    '''

    lines = pysam.idxstats(bamfile).splitlines()

    try:
        nreads = sum(
            map(int, [x.split("\t")[2]
                      for x in lines if not x.startswith("#")]))

    except IndexError as msg:
        raise IndexError(
            "can't get number of reads from bamfile, msg=%s, data=%s" %
            (msg, lines))
    return nreads


def isStripped(bamfile):
    '''
    Check if the sequence is stripped in a bam file.
    '''
    bam = pysam.AlignmentFile(bamfile)
    read = bam.next()
    if read.seq is None:
        return True
    else:
        return False


def merge_pairs(AlignmentFile input_samfile,
                outfile,
                min_insert_size = 0,
                max_insert_size = 400,
                bed_format = None ):
    '''merge paired ended data.

    For speed reasons, the aligned region is only approximated using
    the read length. Indels within reads are ignored. Thus bed coordinates
    might be a few residues off.

    The strand is always set to '+'.

    Pairs with a maximum insert size larger than *max_insert_size* are removed.

    If `bed_format` is a number, only the first x columns will be output.
    '''

    cdef int ninput = 0
    cdef int nremoved_unmapped = 0
    cdef int nremoved_insert = 0
    cdef int nremoved_contig = 0
    cdef int nremoved_unpaired = 0
    cdef int nremoved_coordinate = 0
    cdef int nremoved_take_only_second = 0
    cdef int noutput = 0
    cdef int flag
    cdef int isize

    cdef AlignedSegment read
    cdef int c_max_insert_size = max_insert_size
    cdef int c_min_insert_size = min_insert_size
    cdef int start, end, xstart, xend
    cdef int take_columns = 6

    # point to array of contig lengths
    cdef uint32_t *contig_sizes = input_samfile.header.ptr.target_len
    
    if bed_format != None:
        if bed_format < 3 or bed_format > 6: 
            raise ValueError("a bed file must have at least 3 and at most 6 columns")
        take_columns = bed_format

    for read in input_samfile:
        ninput += 1

        flag = read._delegate.core.flag 
        # remove unmapped reads
        if flag & 4:
            nremoved_unmapped += 1
            continue

        if read.pos < read.mpos:
            # lower coordinate than mate, ignore
            nremoved_coordinate += 1
            continue
        elif read.pos == read.mpos and flag & 64:
            # disambiguate, ignore first in pair
            nremoved_take_only_second += 1
            continue
        else:
            # taking the downstream pair allows to incl
            xstart = read.next_reference_start
            xend = read.reference_end
            if xstart < xend:
                start, end = xstart, xend
            else:
                start, end = xend, xstart

        # remove unpaired
        if not flag & 2:
            nremoved_unpaired += 1
            continue
            
        if read.tid != read.mrnm:
            nremoved_contig += 1
            continue

        # isize can be negative - depending on the pair orientation
        isize = abs(read.isize)
        if (c_max_insert_size and isize > c_max_insert_size) or \
           (c_min_insert_size and isize < c_min_insert_size):
            nremoved_insert += 1
            continue

        # truncate at contig end - overhanging reads might cause problems with chrM
        if end > contig_sizes[read.mrnm]:
            end = contig_sizes[read.mrnm]

        # count output pair as two so that it squares with ninput
        noutput += 2

        if take_columns == 3:
            outfile.write("%s\t%i\t%i\n" %
                          (input_samfile.getrname(read.tid),
                           start, end))
        elif take_columns == 4:
            outfile.write("%s\t%i\t%i\t%s\n" % 
                          (input_samfile.getrname(read.tid),
                           start, end,
                           read.qname))
        elif take_columns == 5:
            outfile.write("%s\t%i\t%i\t%s\t%i\n" % 
                          (input_samfile.getrname(read.tid),
                           start, end,
                           read.qname,
                           read.mapq,
                       ))
        else:
            # As we output the downstream read, reverse orientation
            if read.is_reverse:
                strand = '-'
            else:
                strand = '+'
            outfile.write("%s\t%i\t%i\t%s\t%i\t%s\n" % 
                          (input_samfile.getrname(read.tid),
                           start, end,
                           read.qname,
                           read.mapq,
                           strand
                          ))

    c = E.Counter()
    c.input = ninput
    c.removed_insert = nremoved_insert
    c.removed_contig = nremoved_contig
    c.removed_unmapped = nremoved_unmapped
    c.removed_unpaired = nremoved_unpaired
    c.removed_take_only_second = nremoved_take_only_second
    c.removed_coordinate = nremoved_coordinate
    c.output = noutput

    return c


def bams2bam_filter(AlignmentFile genome_samfile,
                    AlignmentFile output_samfile,
                    AlignmentFile output_mismapped,
                    AlignmentFile transcripts_samfile,
                    AlignmentFile junctions_samfile,
                    transcripts,
                    regions = None,
                    unique = False,
                    remove_contigs = None,
                    colour_mismatches = False,
                    ignore_mismatches = False,
                    ignore_junctions = True,
                    ignore_transcripts = False ):
    '''
    To conserve memory, the tid and NM flag from *transcripts_samfile*
    are packed into memory. As a consequence, this method requires 
    max(NM) < 256 (2^8) and max(tid) < 16777216 (2^24)

    If *unique* is set, only uniquely matching reads will be output.

    If *remove_contigs* is given, contigs that are in remove_contigs will
    be removed. Note that this will only remove the alignment, but not
    all alignments of a particuluar read and the NH flag will *NOT* be
    updated.

    If *colour_mismatches* is set, the ``CM`` tag will be used
    to count differences. By default, the ``NM`` tag is used.
    The tag that is used needs to present in both *transcripts_samfile*
    and *genome_samfile*.

    If *ignore_mismatches* is set, the number of mismatches is ignored.

    If *regions* is given, alignments overlapping regions will be removed.

    '''
    cdef int ninput = 0
    cdef int nunmapped_genome = 0
    cdef int nunmapped_transcript = 0
    cdef int nmismapped = 0
    cdef int noutput = 0
    cdef int nnotfound = 0
    cdef int nlocation_ok = 0
    cdef int nnonunique = 0
    cdef int ntested = 0
    cdef int nremoved_contigs = 0
    cdef int nremoved_junctions = 0
    cdef int nskipped_junctions = 0
    cdef int nadded_junctions = 0
    cdef int nremoved_regions = 0

    cdef bint c_unique = unique
    cdef bint c_test_mismatches = not ignore_mismatches
    cdef bint c_test_junctions = not ignore_junctions
    cdef bint c_test_transcripts = not ignore_transcripts
    cdef bint c_test_regions = regions
    cdef int * remove_contig_tids 
    cdef int nremove_contig_tids = 0
    cdef AlignedSegment read
    cdef AlignedSegment match
    
    # build index
    # this method will start indexing from the current file position
    # if you decide
    cdef int ret = 1
    cdef int x
    cdef bam1_t * b = <bam1_t*>calloc(1, sizeof( bam1_t))
    cdef uint64_t pos
    cdef uint8_t * v
    cdef int32_t nm
    cdef int32_t nh
    cdef int tid
    cdef int transript_tid
    cdef long val
    cdef bint skip = 0
    cdef int * map_tid2tid

    cdef int32_t read_mismatches = 0

    # decide which tag to use
    cdef char * nm_tag = 'NM' 
    cdef char * cm_tag = 'CM'
    cdef char * tag

    if colour_mismatches:
        tag = cm_tag
    else:
        tag = nm_tag

    # set with junctions that are ignored
    skip_junctions = set()

    # build list of junctions
    if c_test_junctions:
        E.info("building junction read index")

        def _gen2(): return array.array('B') 
        junctions_index = collections.defaultdict(_gen2)
        ret = 1
        while ret > 0:
            ret = bam_read1(hts_get_bgzfp(junctions_samfile.htsfile),
                            b)
            if ret > 0:
                # ignore unmapped reads
                if b.core.flag & 4: continue
                qname = pysam_bam_get_qname(b)
                v = bam_aux_get(b, tag)
                nm = <int32_t>bam_aux2i(v)
                junctions_index[qname].append(nm)

        E.info( "built index for %i junction reads" % len(junctions_index))

        map_tid2tid = <int*>calloc(len(junctions_samfile.references), sizeof(int))
        
        for x, contig_name in enumerate(junctions_samfile.references):
            map_tid2tid[x] = genome_samfile.gettid(contig_name)

    if c_test_transcripts:
        E.info( "building transcriptome read index" )
        # L = 4 bytes
        def _gen(): return array.array('L') 
        transcriptome_index = collections.defaultdict(_gen)
        ret = 1
        while ret > 0:
            ret = bam_read1(hts_get_bgzfp(transcripts_samfile.htsfile),
                            b)
            if ret > 0:
                # ignore unmapped reads
                if b.core.flag & 4: continue
                qname = pysam_bam_get_qname(b)
                tid = b.core.tid
                v = bam_aux_get(b, tag)
                nm = <int32_t>bam_aux2i(v)
                transcriptome_index[qname].append( (tid << 8) | nm )

        E.info( "built index for %i transcriptome reads" % len(transcriptome_index))

    bam_destroy1( b )

    # setup list of contigs to remove:
    if remove_contigs:
        nremove_contig_tids = len(remove_contigs)
        remove_contig_tids = <int*>malloc( sizeof(int) * nremove_contig_tids )
        for x, rname in enumerate( remove_contigs):
            remove_contig_tids[x] = genome_samfile.gettid( rname )
    
    E.info( "starting filtering" )

    for read in genome_samfile:

        ninput += 1
        # if ninput > 10000: break

        # is unmapped?
        if read._delegate.core.flag & 4:
            nunmapped_genome += 1
            noutput += 1
            output_samfile.write( read )
            continue

        # optionally remove non-unique reads
        if c_unique:
            v = bam_aux_get(read._delegate, 'NH')
            if v != NULL:
                nh = <int32_t>bam_aux2i(v)
                if nh > 1: 
                    nnonunique += 1
                    continue

        # optionally remove reads matched to certain contigs
        if nremove_contig_tids:
            skip = 0
            for x from 0 <= x < nremove_contig_tids:
                if remove_contig_tids[x] == read.tid:
                    nremoved_contigs += 1
                    skip = 1
                    break
            if skip: continue

        g_contig = genome_samfile.getrname( read.tid )
                
        # optionally remove reads mapped to certain regions
        if c_test_regions:
            intervals = regions.get( g_contig, read.pos, read.aend )
            skip = 0
            for start, end, value in intervals:
                if read.overlap( start, end ):
                    nremoved_regions += 1
                    skip = 1
                    break

            if skip: continue

        if c_test_junctions:
            if read.qname in junctions_index:
                # can compress index before, depends on
                # how many reads you expect to test
                nm = min(junctions_index[read.qname])
                read_mismatches = read.opt(tag)
                if nm > read_mismatches: 
                    nremoved_junctions += 1
                    continue
                else:
                    skip_junctions.add(read.qname)
                                        
        # set mapped = True, if read is mapped to transcripts
        # set location_ok = True, if read matches in expected location
        # according to transcripts
        location_ok = False
        mapped = False

        if c_test_transcripts:
            # get transcripts that read matches to
            try:
                matches = transcriptome_index[read.qname]
            except KeyError:
                nnotfound += 1
                matches = None
                
            if matches:

                if c_test_mismatches:
                    read_mismatches = read.opt(tag)

                for val in matches:
                    transcript_tid = val >> 8
                    # ignore reads that are mapped to transcripts with
                    # more mismatches than the genomic location
                    if c_test_mismatches:
                        nm = val & 255
                        if nm > read_mismatches: continue

                    mapped = True

                    # find transcript
                    transcript_id = transcripts_samfile._getrname( transcript_tid )
                    gtfs = transcripts[transcript_id]
                    t_contig, t_start, t_end = gtfs[0].contig, gtfs[0].start, gtfs[-1].end

                    # does read map to genomic transcript location?
                    if g_contig == t_contig and t_start <= read.pos <= t_end:
                        location_ok = True
                        break
                    
        if location_ok:
            ntested += 1
            nlocation_ok += 1
            noutput += 1
            output_samfile.write( read )

        elif mapped:
            ntested += 1
            nmismapped += 1
            if output_mismapped:
                output_mismapped.write( read )

        else:
            nunmapped_transcript += 1
            noutput += 1
            output_samfile.write( read )

    # add junctions
    if c_test_junctions:
        E.info("adding junctions")
        junctions_samfile.reset()

        # rebuild contig map for junctions:
        if remove_contigs:
            remove_contig_tids = <int*>malloc( sizeof(int) * nremove_contig_tids )
            for x, rname in enumerate( remove_contigs):
                remove_contig_tids[x] = junctions_samfile.gettid( rname )
        
        for read in junctions_samfile:

            # optionally remove reads matched to certain contigs
            if nremove_contig_tids:
                skip = 0
                for x from 0 <= x < nremove_contig_tids:
                    if remove_contig_tids[x] == read.tid:
                        nremoved_contigs += 1
                        skip = 1
                    break
                if skip: 
                    nskipped_junctions += 1
                    continue

            if read.qname in skip_junctions:
                nskipped_junctions += 1
            else:
                nadded_junctions += 1
                noutput += 1
                # map tid from junction database to genome database
                read.tid = map_tid2tid[read.tid]
                output_samfile.write(read)
        
        free( map_tid2tid )

    c = E.Counter()
    c.input = ninput
    c.removed_nonunique = nnonunique
    c.removed_mismapped = nmismapped
    c.removed_contigs = nremoved_contigs
    c.removed_junctions = nremoved_junctions
    c.removed_regions = nremoved_regions
    c.skipped_junction_reads = len(skip_junctions)
    c.skipped_junctions = nskipped_junctions
    c.added_junctions = nadded_junctions
    c.output = noutput
    c.unmapped_genome = nunmapped_genome
    c.unmapped_transcript = nunmapped_transcript
    c.notfound = nnotfound
    c.location_ok = nlocation_ok 
    c.location_tested = ntested

    if nremove_contig_tids:
        free( remove_contig_tids )

    E.info( "filtering finished" )

    return c



"""convert fasta to VCF
==========================

Output a file in VCF format with variants according
to a fasta file.

Note: currently only takes first sample in VCF and assumes
single-sample VCF

"""

import sys
import random
import collections
import cgatcore.experiment as E
import numpy
import pysam


class Counter(object):
    pass


class CounterIndelType(Counter):

    header = ["indel_type",
              "indel_length",
              "indel_delta",
              "indel_class"]

    def count(self, record):
        self.ignore = False
        self.indel_length = ""
        self.indel_class = ""
        self.indel_type = ""
        self.indel_delta = ""
        ref = record.ref
        alt = record.alts[0]
        if len(record.alts) > 1:
            return
        lref = len(ref)
        lalt = len(alt)
        if lref == lalt:
            return
        elif lref > lalt:
            self.indel_type = "deletion"
            self.indel_delta = ref[1:]
        else:
            self.indel_type = "insertion"
            self.indel_delta = alt[1:]

        self.indel_length = len(self.indel_delta)
        counts = collections.Counter(self.indel_delta)
        if len(counts) == 1 and self.indel_length > 2:
            self.indel_class = "mononucleotide-instability"
        elif len(counts) == 2 and self.indel_length > 4:
            self.indel_class = "dinucleotide-instability"
        else:
            self.indel_class = "other"

    def __str__(self):
        return "\t".join(map(
            str,
            (self.indel_type, self.indel_length,
             self.indel_delta, self.indel_class)))


class CounterContext(Counter):

    header = ["class", "left_class", "right_class", "context"]

    def __init__(self, fasta):

        if fasta is None:
            raise ValueError("CounterContext requires an indexed fasta file")
        self.fasta = fasta

        self.region = 20

    def count(self, record):

        pos = record.pos - 1
        s = self.fasta.fetch(record.chrom,
                             pos - self.region,
                             pos + self.region + 1)

        def classify(region, threshold_low_complexity_region=5, threshold_homopolymer=10):
            c = collections.Counter(region)

            # monomers define homopolymers
            if len(list(c.values())) == 0:
                return "empty"

            if max(c.values()) > threshold_homopolymer:
                return "hompolymer"

            # dimers define low-complexity regions
            c = collections.Counter([region[x:x+2] for x in range(len(region)-1)])
            if len(list(c.values())) == 0:
                return "empty"

            if max(c.values()) > threshold_low_complexity_region:
                return "low-complexity"

            return "-"

        left_region = s[:self.region]
        right_region = s[self.region + 1:]
        self.cls_left = classify(left_region)
        self.cls_right = classify(right_region)

        self.context = left_region.lower() + s[self.region] + right_region.lower()

        full_cls = (self.cls_left, self.cls_right)

        if self.cls_left == self.cls_right:
            self.cls = self.cls_left
            return

        if "-" in full_cls:
            self.cls = "border"
        else:
            self.cls = "repetetive"

    def __str__(self):
        return "\t".join((self.cls, self.cls_left, self.cls_right, self.context))


class CounterBAM(Counter):

    def __init__(self, bam):
        if bam is None:
            raise ValueError("CounterBAM requires an indexed BAM file")

        self.bam = bam


class CounterBAMIndels(CounterBAM):

    header = ["mean_depth",
              "mean_freq_deletions",
              "mean_freq_insertions",
              "mean_freq_indels",
              "positions_with_deletions",
              "positions_with_insertions",
              "positions with_indels",
              "indel_status_string"]

    def __init__(self, *args, **kwargs):
        CounterBAM.__init__(self, *args, **kwargs)

        self.region = 5

    def count(self, record):

        pos = record.pos - 1
        varset = None
        depths = []
        deletions = []
        insertions = []
        for column in self.bam.pileup(record.chrom,
                                      pos - self.region,
                                      pos + self.region + 1,
                                      truncate=True,
                                      stepper="all"):
            depths.append(len(column.pileups))
            deletions.append(sum([x.is_del == 1 for x in column.pileups]))
            insertions.append(sum([x.indel > 0 for x in column.pileups]))

        self.depths = numpy.array(depths)
        self.deletions = numpy.array(deletions, dtype=numpy.float)
        self.insertions = numpy.array(insertions, dtype=numpy.float)

    def __str__(self):

        indels = self.deletions + self.insertions
        indel_status = numpy.floor(numpy.true_divide(indels, self.depths) * 10.0)
        indel_status_string = "".join(map(str, list(map(int, indel_status))))
        return "\t".join(map(str, (
                "{:.2f}".format(numpy.mean(self.depths)),
                "{:.4f}".format(numpy.mean(self.deletions / self.depths)),
                "{:.4f}".format(numpy.mean(self.insertions / self.depths)),
                "{:.4f}".format(numpy.mean(indels / self.depths)),
                sum(self.deletions > 0),
                sum(self.insertions > 0),
                sum(indels > 0),
                indel_status_string,
                )))


class CounterBAMAllelicDepth(CounterBAM):

    header = ["nbases", "ngaps",
              "nref", "nalt", "nother",
              "freq_gaps",
              "freq_ref", "freq_alt", "freq_other",
              "genotype"]

    def __init__(self, *args, **kwargs):
        CounterBAM.__init__(self, *args, **kwargs)

        self.region = 5

    def count(self, record):

        pos = record.pos - 1
        varset = None
        depths = []
        deletions = []
        insertions = []
        for column in self.bam.pileup(record.chrom,
                                      pos,
                                      pos+1,
                                      truncate=True,
                                      stepper="all"):

            bases = [x.alignment.query_sequence[x.query_position]
                     for x in column.pileups if x.query_position is not None]
            self.ngaps = len([x for x in column.pileups if x.query_position is None])
            self.nbases = len(bases)
            self.nreference = len([x for x in bases if x == record.ref])
            if len(record.alts) > 1:
                self.is_multiallelic = True
            alt = record.alts[0]
            self.nalt = len([x for x in bases if x == alt])
            self.nother = self.nbases - self.nreference - self.nalt
            gt = sorted(set(record.samples[0]["GT"]))
            if len(gt) == 1:
                if gt[0] == 1:
                    self.genotype = "hom-alt"
                elif gt[0] == 0:
                    self.genotype = "hom-ref"
                else:
                    self.genotype = "hom-other"
            else:
                self.genotype = "het"

    def __str__(self):
        if self.nbases > 0:
            f_ref = "{:.4f}".format(float(self.nreference) / self.nbases)
            f_alt = "{:.4f}".format(float(self.nalt) / self.nbases)
            f_other = "{:.4f}".format(float(self.nother) / self.nbases)
        else:
            f_ref = f_alt = f_other = "na"

        if self.ngaps + self.nbases > 0:
            f_gaps = "{:.4f}".format(float(self.ngaps) / (self.nbases + self.ngaps))
        else:
            f_gaps = "na"

        return "\t".join(map(str, (
                    self.nbases,
                    self.ngaps,
                    self.nreference,
                    self.nalt,
                    self.nother,
                    f_gaps,
                    f_ref, f_alt, f_other,
                    self.genotype
                    )))


def main(argv=None):

    parser = E.OptionParser(version="%prog version: $Id$",
                            usage=globals()["__doc__"])

    parser.add_option(
        "-s", "--sample-size", dest="sample_size", type="float",
        help="sample size. If less than 0, take a proportion of the chromosome size. "
        "If greater than 0, take a fixed number of variants [%default]")

    parser.add_option(
        "--input-filename-fasta", dest="input_filename_fasta", type="string",
        help="filename with reference sequence in fasta format [%default]")

    parser.add_option(
        "--input-filename-bam", dest="input_filename_bam", type="string",
        help="filename with aligned reads [%default]")

    parser.add_option(
        "--no-vcf-columns", dest="no_vcf_columns", action="store_true",
        help="do not output vcf columns")

    parser.add_option(
        "--counter", dest="counters", type="choice", action="append",
        choices=["context", "bam-indels", "bam-allelic-depth", "indel-type"],
        help="counters to apply [%default]")

    parser.set_defaults(
        input_filename_fasta=None,
        input_filename_bam=None,
        input_filename_vcf=None,
        sample_size=0.001,
        sample_name="NA12878",
        region_size=20,
        threshold_homopolymer=12,
        threshold_repeat=5,
        no_vcf_columns=False,
        counters=[],
    )

    (options, args) = E.start(parser,
                              argv=argv,
                              add_output_options=True)

    if len(args) > 0:
        options.input_filename_vcf = args[0]

    vcf_in = pysam.VariantFile(options.input_filename_vcf)

    counters = []

    if options.input_filename_fasta:
        fasta = pysam.FastaFile(options.input_filename_fasta)
    else:
        fasta = None

    if options.input_filename_bam:
        bam = pysam.AlignmentFile(options.input_filename_bam)
    else:
        bam = None

    for counter in options.counters:
        if counter == "context":
            counters.append(CounterContext(fasta))
        elif counter == "bam-indels":
            counters.append(CounterBAMIndels(bam))
        elif counter == "bam-allelic-depth":
            counters.append(CounterBAMAllelicDepth(bam))
        elif counter == "indel-type":
            counters.append(CounterIndelType())

    outf = options.stdout
    if not options.no_vcf_columns:
        header = str(vcf_in.header).strip().split("\n")[-1].strip()[1:].split("\t")

    else:
        header = ["chrom", "pos"]

    outf.write("\t".join(header))

    for counter in counters:
        outf.write("\t" + "\t".join(counter.header))

    outf.write("\n")
    for record in vcf_in:

        for counter in counters:
            counter.count(record)

        if not options.no_vcf_columns:
            outf.write("{}\t".format(
                str(record).strip()))
        else:
            outf.write("{}\t{}\t".format(
                record.chrom, record.pos))

        outf.write("\t".join(map(str, counters)) + "\n")

    E.stop()


if __name__ == "__main__":
    sys.exit(main())



'''
diamond2counts.py - count alignments to reference
===================================================

:Tags: Python

Purpose
-------

Count the number of alignments to each reference in outfmt6.

Counts are based on various options specified by --method.

best       This will take the best alignment as judged by the highest
           bitscore.




TODO::
Add additional options

Usage
-----

Example::

   python diamond2counts.py

Type::

   python diamond2counts.py --help

for command line help.

Command line options
--------------------

'''

import sys

import cgatcore.experiment as E
from cgat.Diamond import *
import collections
import cgatcore.iotools as iotools


def readCogMap(cog_map):
    '''
    return a dictionary mapping gene to cog
    '''
    gene2cog = {}
    for line in iotools.open_file(cog_map):
        data = line[:-1].split("\t")
        gene2cog[data[0]] = data[1]
    return gene2cog


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if argv is None:
        argv = sys.argv

    # setup command line parser
    parser = E.OptionParser(version="%prog version: $Id$",
                            usage=globals()["__doc__"])

    parser.add_option("-m", "--method", dest="method", type="choice",
                      choices=("best", None),
                      help="method for determing what to count")

    parser.add_option("--sum-cog", dest="sum_cog", action="store_true",
                      help="sum counts over functions (COGs) in --cog-map")

    parser.add_option("--evaluate-cog",
                      dest="evaluate_cog",
                      action="store_true",
                      help="""output the percent of
                              alignments for each read = best hit""")

    parser.add_option("--cog-map", dest="cog_map", type="string",
                      help="file with gene to cog map")

    parser.add_option("-n", "--nsamples", dest="nsamples", type="int",
                      help="""number of queries to evaluate-
                              will take the first n in the file""")

    parser.set_defaults(method=None,
                        sum_cog=False,
                        evaluate_cog=False,
                        cog_map=None,
                        nsamples=10000)

    # add common options (-h/--help, ...) and parse command line
    (options, args) = E.start(parser, argv=argv)

    if options.evaluate_cog:
        assert options.cog_map, """must specify an annotation
                                   mapping gene to function (COG)"""
        assert not options.method, """evaluation performed
                                      in the absence of counting"""

        E.info("reading gene to function (COG) map %s" % options.cog_map)
        gene2cog = readCogMap(options.cog_map)
        E.info("loaded gene to function (COG) map")

        E.info("retrieving alignment data")
        options.stdout.write("query\tpbest\tnalignments\n")

        c = 0
        for alignments in query_iterator(alignment_iterator(options.stdin)):
            c += 1
            scores = []
            if c <= options.nsamples:
                for alignment in alignments:
                    scores.append(alignment.score)
                    best = max(scores)
                    best_alignments = [
                        x for x in alignments if x.score == best]
                    if len(best_alignments) > 1:
                        best_alignments = random.sample(best_alignments, 1)
                    best_alignment = best_alignments[0]
                    best_cog = gene2cog[best_alignment.ref]
                pbest = float(len(
                    [gene2cog[x.ref]
                     for x in alignments
                     if gene2cog[x.ref] == best_cog])) / len(alignments) * 100
                nalignments = len(alignments)
                options.stdout.write(
                    "\t".join(map(
                        str, [alignments[0].qid,
                              pbest,
                              nalignments])) + "\n"
                )
            else:
                break
        return

    # container for counts
    counts = collections.defaultdict(int)
    E.info("counting alignments")
    assert options.method, "required option --method"
    if options.method == "best":
        if options.sum_cog:
            E.warn("""summing over functions (COGS)
                      will remove genes with no annotations
                      and those with multiple COG assignments""")
            assert options.cog_map, """a mapping between gene and
                                       function (COG) is required"""

            E.info("""reading gene to function (COG) mapping from %s"""
                   % options.cog_map)
            gene2cog = readCogMap(options.cog_map)
            E.info("loaded gene to function (COG) mapping")

            E.info("summing functional assignments")
            query_it = query_iterator(alignment_iterator(options.stdin))
            for best in best_alignment_iterator(query_it):
                cog = gene2cog[best.ref]
                # removing uassigned or multiple assignments
                if cog == "unknown" or cog.find(";") != -1:
                    continue
                counts[cog] += 1
        else:
            E.info("counting best alignments")
            query_it = query_iterator(alignment_iterator(options.stdin))
            for best in best_alignment_iterator(query_it):
                counts[best.ref] += 1
        E.info("finished counting")

        E.info("writing results")
        options.stdout.write("ref\tcount\n")
        for ref, count in sorted(counts.items()):
            options.stdout.write("\t".join([ref, str(count)]) + "\n")

    # write footer and output benchmark information.
    E.stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''
cgat_pep8_check_code_quality.py - check PEP8 conformance of cgat Code
=====================================================================

:Author:
:Tags: Python

Purpose
-------

This script runs pep8.py on the cgat code collection and outputs
summary statistics of code quality onto stdout.

Usage
-----

To use, simply run the script from the root directory of the
cgat code collection::

   python cgat_pep8_check_code_quality.py

Type::

   python cgat_pep8_check_code_quality.py --help

for command line help.

Command line options
--------------------


'''

import collections
import sys
import cgatcore.experiment as E
import cgat.Style

DATA = collections.namedtuple('DATA', 'count code description')

expressions = (
    ('tests', 'tests/*.py'),
    ('scripts', 'scripts/*.py'),
    ('optic', 'scripts/optic/*.py'),
    ('gpipe', 'scripts/gpipe/*.py'),
    ('cgat', 'cgat/*.py'),
    ('cgatPipelines', 'cgatPipelines/*.py'),
    ('trackers', 'cgatPipelines/pipeline_docs/*/trackers/*.py'))


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if argv is None:
        argv = sys.argv

    # setup command line parser
    parser = E.OptionParser(version="%prog version: $Id$",
                            usage=globals()["__doc__"])

    # add common options (-h/--help, ...) and parse command line
    (options, args) = E.start(parser, argv=argv)

    rows = []
    labels = {}
    for label, expr in expressions:
        nchecked, data = cgat.Style.runPep8(expr)
        rows.append((label, nchecked, data))
        labels.update(dict([(x.code, x.description) for x in data]))

    # build table
    #
    # each row is data set and each column is a Warning/Error type
    # with some additional columns such as total and n.

    # build dictionary mapping error codes to columns
    # consistently across samples
    map_code2column = dict([(y, x + 3) for x, y in enumerate(labels.keys())])

    # build first row containing the column labels
    results = [['code', 'n', 'total'] + list(labels.keys())]

    # build array with column totals
    column_totals = [0] * (len(map_code2column) + 3)
    for label, nchecked, data in rows:
        row = [label, nchecked, 0] + [0] * len(map_code2column)
        column_totals[1] += nchecked
        for x in data:
            c = map_code2column[x.code]
            row[c] = x.count
            row[2] += int(x.count)
            column_totals[2] += int(x.count)
            column_totals[c] += int(x.count)

        results.append(row)
    # add column totals
    column_totals[0] = 'total'
    results.append(column_totals)

    # add descriptions as last row
    results.append(['description',
                    'number of files checked',
                    'total errors/warnings in set'] + list(labels.values()))

    # output transposed table
    outfile = sys.stdout
    for row in zip(*results):
        outfile.write('%s\n' % ('\t'.join(map(str, row))))

    # write footer and output benchmark information.
    E.stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



import matplotlib
import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy
import os
import pandas
import seaborn

import cgatcore.experiment as E


class MutationProfileBarPlot(object):
    """plot a mutation profile bar-plot.

    dataframe requires the following columns:

    signature:
    context:
    percent_sample
    """

    def __call__(self, dataframe):

        assert "signature" in dataframe.columns
        assert "context" in dataframe.columns
        assert "percent_sample" in dataframe.columns

        current_palette = seaborn.color_palette()
        factors = dataframe.signature.unique()
        map_signature2factor = dict([(y, x) for x, y in
                                     enumerate(factors)])

        colors = [current_palette[map_signature2factor[x]]
                  for x in dataframe.signature]

        handles = []
        for x, factor in enumerate(factors):
            handles.append(
                matplotlib.patches.Patch(color=current_palette[x],
                                         label=factor))

        ax = dataframe.plot(kind="bar",
                            x="context",
                            y="percent_sample",
                            color=colors,
                            legend=False)
        ax.legend(handles=handles)

        ax.figure.set_size_inches(30, 10)
        ax.set_xticklabels([x.get_text() for x in ax.get_xticklabels()],
                           family="monospace")

        return ax


class DepthProfilePlot(object):
    """
    """

    def __call__(self, dataframe, map_sample2label={}):

        df = dataframe.pivot(
            index="gc_bin",
            columns="sample",
            values="mean").reset_index()

        # remove duplicate sample names
        # (in Cancer analysis: two blood samples)
        to_drop = [x for x in df.columns if x.startswith("2:")]
        df.drop(to_drop, axis=1, inplace=True)

        df.colums = [map_sample2label.get(x, x) for x in df.columns]

        if df.empty:
            E.warn("no data, no plot will be output")
            return

        ax = df.plot(kind="line", x="gc_bin")

        return ax


class ChromosomeDataFromFile(object):
    """create melted data frame

    Table format is:

    CHROM   POS   sample1   sample2    ...
    """

    def __call__(self, track, value_label):
        fn = track

        table = pandas.read_csv(fn,
                                comment="#",
                                sep="\t",
                                dtype={"CHROM": object},
                                )

        table = pandas.melt(table, id_vars=["CHROM", "POS"])
        table.colums = ["CHROM", "POS", "sample", value_label]

        try:
            skip = table["depth"].str.startswith(".").fillna(False)
            table = table[~skip]
        except AttributeError:
            pass
        return table.set_index(["CHROM", "POS"])


class CoverageRatioFromFile(object):

    min_depth = 10

    def __call__(self, track, column_is_norm=None, *args, **kwargs):

        fn = track

        table = pandas.read_csv(fn,
                                comment="#",
                                sep="\t",
                                dtype={"CHROM": object},
                                ).set_index(["CHROM", "POS"])

        columns = table.columns

        if len(columns) > 2:
            E.warn("too many columns {}".format(columns))

        if column_is_norm:
            if len(columns) == 2:
                num, den = columns
                if num in column_is_norm:
                    num, den = den, num
            else:
                den = [x for x in columns if x in column_is_norm][0]
                num = [x for x in columns if x not in column_is_norm][0]

            if den not in column_is_norm:
                raise ValueError("denominator is {}, but not norm, cols = {}".format(
                    den, columns))

        table = table[(table[num] >= self.min_depth) &
                      (table[den] >= self.min_depth)]
        for column in columns:
            table[column] = table[column] / table[column].median()

        table["sample"] = num
        table["copy number"] = 2.0 * table[num] / table[den]
        table.drop(columns, axis=1, inplace=True)
        return table


class ChromosomePlot(object):

    position = None
    column = None
    plot_type = None
    window_size = 100
    rolling = None
    subplots = False
    plot_gaps = True

    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, dataframe, chromosome):

        if self.position is None or self.column is None:
            raise NotImplementedError(
                "missing .position or .column attributes")

        plt.figure()
        ax = plt.gca()
        colors = seaborn.color_palette()
        data = dataframe.reset_index()
        labels, plots = [], []
        x = 0
        grouper = data.groupby("track")

        if self.plot_gaps:
            gaps_fn = PARAMS.get("gaps_file", None)
            if gaps_fn is None or gaps_fn == "None":
                gaps = None
            else:
                gaps = pandas.read_csv(gaps_fn,
                                       comment="#",
                                       sep="\t",
                                       header=None)
                gaps = gaps[gaps[0] == chromosome][[0, 1, 2]]

        for name, group in grouper:
            if self.subplots:
                plt.subplot(len(list(grouper.groups.keys())), 1, x + 1)

            if self.plot_type == "scatter":
                plots.append(plt.scatter(group[self.position],
                                         group[self.column],
                                         s=20,
                                         c=colors[x % len(colors)]))
            elif self.plot_type == "line":
                plots.append(plt.plot(group[self.position],
                                      group[self.column],
                                      c=colors[x % len(colors)])[0])
            labels.append(name)
            if self.rolling == "average":
                mean = pandas.rolling_mean(group[self.column],
                                           self.window_size)
                plt.plot(group[self.position],
                         mean,
                         c=colors[x % len(colors)])

            x += 1
            plt.xlim(0, max(data[self.position]))

            plt.tick_params(axis='both', which='major', labelsize=20)
            plt.tick_params(axis='both', which='minor', labelsize=20)

            if gaps is not None:
                ax = plt.gca()
                ymin, ymax = ax.get_ylim()
                for i, row in gaps.iterrows():
                    ax.add_patch(
                        patches.Rectangle(
                            (row[1], ymin),
                            row[2] - row[1],
                            ymax - ymin,
                            color="black",
                            alpha=0.2))

        if self.subplots:
            plt.subplot(len(list(grouper.groups.keys())), 1, 1)
            plt.legend(
                plots, labels,
                bbox_to_anchor=(0., 1.0, 1., .102),
                loc=3,
                ncol=3,
                mode="expand",
                borderaxespad=0.,
                prop={'size': 24}
            )
        else:
            plt.legend(plots, labels)

        ax.figure.set_size_inches(40, 10 * len(list(grouper.groups.keys())))
        return ax


@E.cached_function
def get_chromosome_sizes(filename, vcf_chromosomes=None):

    if filename.endswith(".fa") or filename.endswith(".fasta"):
        fn = filename + ".fai"
        if not os.path.exists(fn):
            raise OSError("file {} with chromosome sizes does not exist"
                          .format(fn))

        with open(fn) as inf:
            chromosome_size = pandas.read_csv(
                inf,
                sep="\t",
                header=None,
                names=("chromosome", "size", "offset",
                       "lw", "lw2"),
                dtype={"chromosome": str},
                index_col="chromosome").drop(["offset", "lw", "lw2"], 1)

    elif filename.endswith(".bed") or filename.endswith(".bed.gz"):

        if not os.path.exists(filename):
            raise OSError("file {} with exons does not exist"
                          .format(filename))
        chromosome_size = pandas.read_csv(
            filename,
            sep="\t",
            header=None,
            names=("chromosome", "start", "end"),
            dtype={"chromosome": str},
        )
        chromosome_size["size"] = (chromosome_size["end"] -
                                   chromosome_size["start"] + 1)

        chromosome_size = chromosome_size.groupby(
            "chromosome").sum().drop(["start", "end"], 1)

    # intersect with VCF
    if vcf_chromosomes:
        chromosome_size = chromosome_size.loc[vcf_chromosomes]

    return chromosome_size.sort_values("size", ascending=False)


class ManhattanPlot(object):

    def __init__(self, genome_size_file, *args, **kwargs):
        if genome_size_file is None:
            raise ValueError(
                "ManhattanPlot requires a file with contig sizes.")

        if not os.path.exists(genome_size_file):
            raise OSError(
                "filename to retrieve contig sizes from does not exist: {}".format(genome_size_file))

        self.genome_size_file = genome_size_file

    def __call__(self,
                 dataframe,
                 chromosome_column="CHROM",
                 position_column="POS",
                 value_column=None,
                 plot_type="scatter",
                 plot_gaps=None,
                 subplots=True,
                 ylabel="read depth",
                 xlabel="chromosome position",
                 rolling=None,
                 window_size=100,
                 yrange=None):

        if chromosome_column not in dataframe.columns:
            raise ValueError(
                "missing chromosome column {}".format(chromosome_column))

        if position_column not in dataframe.columns:
            raise ValueError(
                "missing position column {}".format(position_column))

        if "sample" not in dataframe.columns:
            data = pandas.melt(
                dataframe,
                id_vars=[chromosome_column, position_column]).reset_index().drop(
                    ["index"], axis=1)

        else:
            if value_column is None:
                raise ValueError("a melted table requires value_column to be provided")

            data = dataframe[[chromosome_column,
                              position_column,
                              "sample",
                              value_column]]

        data.columns = [chromosome_column, position_column,
                        "sample", "value"]

        try:
            skip = data["value"].str.startswith(".").fillna(False)
            data = data[~skip]
        except AttributeError:
            pass

        plt.figure()
        ax = plt.gca()
        colors = seaborn.color_palette()
        labels, plots = [], []

        # build increment for chromosomal positions
        chromosome_sizes = get_chromosome_sizes(self.genome_size_file)
        chromosome_offsets = chromosome_sizes.cumsum().shift(1).fillna(0)
        chromosome_offsets_dict = dict(
            [(str(x), int(y)) for x, y in chromosome_offsets.iterrows()])

        # remove variants on unknown chromosomes
        known_chromosomes = set(chromosome_sizes.index)
        data = data[data[chromosome_column].isin(known_chromosomes)]

        if len(data) == 0:
            raise ValueError("after chromosome filtering, no data remained")

        # sort by order in chromonome sizes
        data[chromosome_column] = pandas.Categorical(
            data[chromosome_column].astype(str),
            chromosome_sizes.index)

        data.sort_values(by=[chromosome_column, position_column],
                         inplace=True)
        data[position_column] = [y + chromosome_offsets_dict[str(x)]
                                 for x, y in zip(data[chromosome_column],
                                                 data[position_column])]

        if plot_gaps:
            if not os.path.exits(plot_gaps):
                raise ValueError("filename {} with gap data not found".format(
                    plot_gaps))
            gaps = pandas.read_csv(plot_gaps,
                                   comment="#",
                                   sep="\t",
                                   header=None)
            gaps.columns = ["contig", "start", "end"]
            gaps = gaps[gaps.contig.isin(chromosome_offsets_dict)]
            gaps.start = [y + chromosome_offsets_dict[str(x)]
                          for x, y in zip(gaps.contig,
                                          gaps.start)]
            gaps.end = [y + chromosome_offsets_dict[str(x)]
                        for x, y in zip(gaps.contig,
                                        gaps.end)]
        else:
            gaps = None

        chromosome_colors = seaborn.color_palette(
            "Paired",
            len(chromosome_offsets))

        column = "value"
        grouper = data.groupby(by="sample")
        subp = 0
        for name, group in grouper:
            if subplots:
                plt.subplot(len(list(grouper.groups.keys())), 1, subp + 1)

            if yrange is not None:
                if yrange.startswith("percentile-"):
                    percentile = float(yrange.split("-")[1])
                    ymin = 0
                    ymax = numpy.percentile(group[column],
                                            percentile)
                elif yrange.startswith("median-"):
                    factor = float(yrange.split("-")[1])
                    ymin = 0
                    ymax = numpy.percentile(group[column],
                                            50.0) * factor
                elif yrange.startswith("max-"):
                    factor = float(yrange.split("-")[1])
                    ymin = 0
                    ymax = numpy.max(group[column]) * factor
                elif yrange.startswith("mean-"):
                    factor = float(yrange.split("-")[1])
                    ymin = 0
                    ymax = numpy.mean(group[column]) * factor
                elif yrange.startswith("balanced-"):
                    ymin = 0
                    # strike balance: truncate at high peaks,
                    # but at least certain multiple of mean
                    factor = float(yrange.split("-")[1])
                    ymax = max(
                        numpy.mean(group[column] * factor),
                        numpy.max(group[column]) * 0.5)
                elif yrange.startswith("threshold-"):
                    ymin = 0
                    # apply threshold, but only if max is below
                    factor = float(yrange.split("-")[1])
                    ymax = min(
                        factor,
                        numpy.max(group[column]))

                # truncate all dots to max value
                group[column][group[column] > ymax] = ymax

            if plot_type == "scatter":
                plots.append(plt.scatter(group[position_column],
                                         group[column],
                                         s=5,
                                         c=colors[subp % len(colors)]))
            elif plot_type == "line":
                plots.append(plt.plot(group[position_column],
                                      group[column],
                                      c=colors[subp % len(colors)])[0])
            labels.append(name)
            if rolling == "average":
                mean = pandas.rolling_mean(group[column],
                                           window_size)
                plt.plot(group[position_column],
                         mean,
                         c=colors[subp % len(colors)])

            subp += 1
            plt.xlim(0, max(data[position_column]))

            plt.tick_params(axis='both', which='major', labelsize=20)
            plt.tick_params(axis='both', which='minor', labelsize=20)

            if yrange is not None:
                ax.set_ylim((ymin, ymax * 1.01))

            ymin, ymax = ax.get_ylim()
            ax = plt.gca()

            if gaps is not None:
                ax = plt.gca()
                ymin, ymax = ax.get_ylim()
                for i, row in gaps.iterrows():
                    ax.add_patch(
                        patches.Rectangle(
                            (row[1], ymin),
                            row[2] - row[1],
                            ymax - ymin,
                            color="black",
                            alpha=0.2))

            last_pos = 0

            for i, row in enumerate(chromosome_offsets.iterrows()):
                label, pos = row
                pos = int(pos)
                ax.add_patch(
                    patches.Rectangle(
                        (last_pos, ymin),
                        pos - last_pos,
                        ymax - ymin,
                        color=chromosome_colors[i],
                        alpha=0.4))
                last_pos = pos

            # set major ticks and turn of labels
            ax.set_xticks(list(chromosome_offsets["size"]))
            ax.xaxis.set_major_formatter(matplotlib.ticker.NullFormatter())

            # label at midpoint
            midpoints = [(a + b) / 2 for a, b in zip(chromosome_offsets["size"][:-1],
                                                     chromosome_offsets["size"][1:])]
            midpoints.append(chromosome_offsets["size"][-1])

            ax.xaxis.set_minor_locator(
                matplotlib.ticker.FixedLocator(midpoints))
            ax.xaxis.set_minor_formatter(matplotlib.ticker.FixedFormatter(
                chromosome_offsets.index))
            ax.set_ylabel(ylabel)

        if subplots:
            plt.subplot(len(list(grouper.groups.keys())), 1, 1)
            plt.legend(
                plots, labels,
                bbox_to_anchor=(0., 1.0, 1., .102),
                loc=3,
                ncol=3,
                mode="expand",
                borderaxespad=0.,
                prop={'size': 24}
            )
        else:
            plt.legend(plots, labels)

        ax.figure.set_size_inches(40, 10 * len(list(grouper.groups.keys())))
        plt.xlabel(xlabel)

        return ax


class ManhattanPlotWithCopyNumberBars(ManhattanPlot):

    def __init__(self, *args, **kwargs):
        ManhattanPlot.__init__(self, *args, **kwargs)

    def __call__(self, filenames_bed=None, *args, **kwargs):

        c = ManhattanPlot()

        if not filenames_bed:
            return c

        # build increment for chromosomal positions
        chromosome_sizes = get_chromosome_sizes(library="genome")
        chromosome_offsets = chromosome_sizes.cumsum().shift(1).fillna(0)
        chromosome_offsets_dict = dict(
            [(str(x), int(y)) for x, y in chromosome_offsets.iterrows()])
        colors = seaborn.color_palette()
        ax = plt.gca()
        row_index = 1
        y_width = 0.1

        for filename_bed in filenames_bed:
            df = pandas.read_csv(filename_bed,
                                 sep="\t",
                                 header=None,
                                 names=["contig", "start", "end", "copynumber"])
            df["start"] = [y + chromosome_offsets_dict[str(x)]
                           for x, y in zip(df.contig, df.start)]
            df["end"] = [y + chromosome_offsets_dict[str(x)]
                         for x, y in zip(df.contig, df.end)]

            colors = colors[row_index % len(colors)]
            for i, row in df.iterrows():
                alpha = 1.0
                y = row.copynumber
                ax.add_patch(
                    patches.Rectancle(
                        (row.start, y),
                        row.end - row.start,
                        y_width,
                        color=colors,
                        alpha=alpha))
            row_index += 1
        ylimits = ax.get_ylim()
        plt.ylim(0, min(10, ylimits[1]))

        return c


class HeterozygosityFrequenciesPlot(ChromosomePlot):
    position = "position"
    column = "density"
    plot_type = "line"
    ylabel = "proportion of heterozygous variants in window"

    def __init__(self, *args, **kwargs):
        ChromosomePlot.__init__(self, *args, **kwargs)

    def __call__(self, dataframe, path):

        c = ChromosomePlot()
        c.position = self.position
        c.column = self.column
        c.plot_type = self.plot_type
        result = c(dataframe, path)

        df = ROHRegions()(path)

        y_width = 0.02
        y_offset = 1.0 - y_width

        colors = seaborn.color_palette()
        x = 0
        ax = plt.gca()
        for key, group in df.groupby("sample"):
            color = colors[x % len(colors)]
            for i, row in group.iterrows():
                if row.state == 0:
                    alpha = 0.1
                else:
                    alpha = 1.0

                ax.add_patch(
                    patches.Rectangle(
                        (row.start, y_offset),
                        row.end - row.start,
                        y_width,
                        color=color,
                        alpha=alpha))

            y_offset -= y_width
            x += 1
        plt.ylim(0, 1)

        return result


class HeterozygosityFrequenciesManhattanPlot(ManhattanPlot):
    position = "position"
    column = "density"
    plot_type = "line"
    ylabel = "proportion of heterozygous variants in window"

    def __init__(self, *args, **kwargs):
        ManhattanPlot.__init__(self, *args, **kwargs)

    def __call__(self, dataframe, path):

        c = ManhattanPlot()
        c.position = self.position
        c.column = self.column
        c.plot_type = self.plot_type
        result = c(dataframe, path)

        df = ROHRegions()()

        # build increment for chromosomal positions
        chromosome_sizes = get_chromosome_sizes(library="genome")
        chromosome_offsets = chromosome_sizes.cumsum().shift(1).fillna(0)
        chromosome_offsets_dict = dict(
            [(str(x), int(y)) for x, y in chromosome_offsets.iterrows()])

        df["start"] = [y + chromosome_offsets_dict[str(x)]
                       for x, y in zip(df.contig,
                                       df.start)]
        df["end"] = [y + chromosome_offsets_dict[str(x)]
                     for x, y in zip(df.contig,
                                     df.end)]

        y_width = 0.02
        y_offset = 1.0 - y_width

        colors = seaborn.color_palette()
        x = 0
        ax = plt.gca()

        for key, group in df.groupby("sample"):
            color = colors[x % len(colors)]
            for i, row in group.iterrows():
                if row.state == 0:
                    alpha = 0.1
                else:
                    alpha = 1.0

                ax.add_patch(
                    patches.Rectangle(
                        (row.start, y_offset),
                        row.end - row.start,
                        y_width,
                        color=color,
                        alpha=alpha))

            y_offset -= y_width
            x += 1
        plt.ylim(0, 1)

        return result



"""Compare alignments in BAM Files
====================================

This script compares the alignments in a :term:`BAM` formatted file
against alignments in a reference (also :term:`BAM` formatted).

The script outputs a summary of missing alignments as well as a detailed
table containing for each read the following information:

.. csv-table::
   :header: "column_name", "description"

   "read", "name of the read"
   "length", "length of the read"
   "status", "status, one of (unique|multi)_(mapped|mismapped)"
   "overlap", "physical overlap between reads"
   "comp_contig", "contig that read maps to in test set",
   "comp_start", "chromosomal position that reads maps to"
   "comp_end", "chromosomal position that reads maps to"
   "ref_contig", "contig that read maps to in reference set"
   "ref_start", "chromosomal position that reads maps to"
   "ref_end", "chromosomal position that reads maps to"
   "shared_misaligned", "misaligned residue pairs between test and reference"
   "shared_aligned", "correctly aligned residue pairs shared between test and reference"
   "shared_insertion", "correctly aligned insertions"
   "shared_deletion", "correctly aligned deletions"
   "comp_aligned", "aligned pairs in test read"
   "comp_insertion", "inserted bases in test read"
   "comp_deletion", "deleted bases in test read"
   "ref_aligned", "aligned pairs in reference read"
   "ref_insertion", "inserted bases in reference read"
   "ref_deletion", "deleted bases in reference read"

.. note::

   This script assumes single-end data (for example, ONT).

"""
import sys
import os
import pysam
import re
import cgatcore.experiment as E


def count_pairs(s):
    insertions = len([x for x in s if x[1] is None])
    deletions = len([x for x in s if x[0] is None])
    paired = len(s) - insertions - deletions
    return paired, insertions, deletions


def iterate_read_pairs(stream1, stream2, qname_fn=None):

    read1 = next(stream1)
    read2 = next(stream2)

    def passthrough(x):
        return x

    if qname_fn is None:
        qname_fn = passthrough

    try:
        n1 = qname_fn(read1.query_name)
        n2 = qname_fn(read2.query_name)
        while 1:
            if n1 == n2:
                yield(read1, read2)
                read1 = next(stream1)
                read2 = next(stream2)
                n1 = qname_fn(read1.query_name)
                n2 = qname_fn(read2.query_name)
            elif n1 < n2:
                yield(read1, None)
                read1 = next(stream1)
                n1 = qname_fn(read1.query_name)
            else:
                yield(None, read2)
                read2 = next(stream2)
                n2 = qname_fn(read2.query_name)
    except StopIteration:
        pass


def group_pairs(stream):

    reads1, reads2 = [], None
    last_name = None

    for read1, read2 in stream:
        if read1 is None:
            name = read2.query_name
            if last_name == name:
                reads2 = read2
            else:
                if reads2 is not None:
                    yield reads1, reads2
                reads1, reads2 = [], read2
                last_name = name
        elif read2 is None:
            name = read1.query_name
            assert last_name == name
            reads1.append(read1)
        else:
            name = read1.query_name
            if name != last_name:
                if reads2 is not None:
                    yield reads1, reads2
                reads1, reads2 = [read1], read2
            else:
                assert False, "this should not happen"
            last_name = name

    yield reads1, reads2


def main(argv=sys.argv):

    parser = E.OptionParser(version="%prog version: $Id$",
                            usage=globals()["__doc__"])

    parser.add_option(
        "-i", "--input-bam", dest="input_bam_file", type="string",
        help="input bam file")

    parser.add_option(
        "-f", "--reference-bam", dest="reference_bam_file", type="string",
        help="reference BAM file [%default]")

    parser.add_option(
        "-q", "--query-name-regex", dest="query_name_regex", type="string",
        help="regular expression to apply on query name. "
        "Potentially required to match samtools sort order and should "
        "evaluate to an integer [%default]")

    parser.set_defaults(
        input_bam_file=None,
        reference_bam_file=None,
        query_name_regex=None,
    )

    (options, args) = E.start(parser, argv, add_output_options=True)

    if len(args) == 2:
        options.input_bam_file = args[0]
        options.reference_bam_file = args[1]

    if options.input_bam_file is None:
        raise ValueError("please supply a BAM file as input")

    if options.reference_bam_file is None:
        raise ValueError("please supply a BAM file as reference")

    # update paths to absolute
    options.input_bam_file = os.path.abspath(options.input_bam_file)
    options.reference_bam_file = os.path.abspath(options.reference_bam_file)

    if not os.path.exists(options.input_bam_file):
        raise OSError("input bam file {} does not exist".format(
            options.input_bam_file))

    if not os.path.exists(options.reference_bam_file):
        raise OSError("reference bam file {} does not exist".format(
            options.reference_bam_file))

    bam_in = pysam.AlignmentFile(options.input_bam_file)
    ref_in = pysam.AlignmentFile(options.reference_bam_file)

    outf_mapped = E.open_output_file("mapped")
    outf_mapped.write("\t".join(
        ["read",
         "length",
         "status",
         "overlap",
         "comp_contig",
         "comp_start",
         "comp_end",
         "ref_contig",
         "ref_start",
         "ref_end",
         "shared_misaligned",
         "shared_aligned",
         "shared_insertion",
         "shared_deletion",
         "comp_aligned",
         "comp_insertion",
         "comp_deletion",
         "ref_aligned",
         "ref_insertion",
         "ref_deletion"]) + "\n")

    outf_missing = E.open_output_file("missing")
    outf_missing.write("\t".join(
        ["read", "length", "status", "aligned",
         "insertion", "deletion"]) + "\n")

    counter = E.Counter()

    if options.query_name_regex:
        rx = re.compile(options.query_name_regex)

    def extract_query(x):
        return int(rx.search(x).groups()[0])

    qname_fn = None
    if options.query_name_regex:
        qname_fn = extract_query

    for reads_cmp, read_ref in group_pairs(iterate_read_pairs(
            bam_in.fetch(until_eof=True),
            ref_in.fetch(until_eof=True),
            qname_fn=qname_fn)):

        if len(reads_cmp) == 0:
            counter.missing += 1
            pairs_ref = set(read_ref.get_aligned_pairs())
            outf_missing.write("\t".join(
                map(str, (
                    read_ref.query_name,
                    read_ref.query_length,
                    "missing") +
                    count_pairs(pairs_ref))) + "\n")
            continue

        if len(reads_cmp) > 1:
            # multiple matches
            counter.multi_mapping += 1
            prefix = "multi_"
        else:
            counter.unique_mapping += 1
            prefix = "unique_"

        is_mapped = False
        for read_cmp in reads_cmp:

            counter.paired += 1

            if read_cmp.is_unmapped:
                counter.unmapped += 1
                pairs_ref = set(read_ref.get_aligned_pairs())
                outf_missing.write("\t".join(
                    map(str, (
                        read_ref.query_name,
                        read_ref.query_length,
                        "unmapped") +
                        count_pairs(pairs_ref))) + "\n")
                continue

            overlap = max(0, (min(read_cmp.reference_end,
                                  read_ref.reference_end) -
                              max(read_cmp.reference_start,
                                  read_ref.reference_start)))

            pairs_cmp = set(read_cmp.get_aligned_pairs())
            pairs_ref = set(read_ref.get_aligned_pairs())
            shared_cmp = pairs_cmp.intersection(pairs_ref)
            unique_cmp = pairs_cmp.difference(pairs_ref)
            missaligned = len([x for x, y in unique_cmp
                               if x is not None and y is not None])

            if read_cmp.reference_name != read_ref.reference_name or \
               overlap == 0:
                status = "mismapped"
            else:
                counter.overlap += 1
                status = "mapped"
                is_mapped = True

            outf_mapped.write("\t".join(
                map(str, (read_cmp.query_name,
                          read_cmp.query_length,
                          prefix + status,
                          overlap,
                          read_cmp.reference_name,
                          read_cmp.reference_start,
                          read_cmp.reference_end,
                          read_ref.reference_name,
                          read_ref.reference_start,
                          read_ref.reference_end,
                          missaligned) +
                    count_pairs(shared_cmp) +
                    count_pairs(pairs_cmp) +
                    count_pairs(pairs_ref))) + "\n")
        else:
            if is_mapped:
                status = "mapped"
            else:
                status = "mismapped"

            counter[prefix + status] += 1

    with E.open_output_file("summary") as outf:
        outf.write("category\tcounts\n")
        outf.write(counter.asTable() + "\n")

    E.stop()

if __name__ == "__main__":
    sys.exit(main())



'''csv2db.py - upload table to database
====================================

:Tags: Python

Purpose
-------

create a table from a csv separated file and load data into it.

This module supports backends for postgres and sqlite3. Column types are
auto-detected.

Read a table from stdin and create an sqlite3 database. By default,
the database will reside in a file called csvdb and in a table csv.

.. todo::

   Use file import where appropriate to speed up loading. Currently, this is
   not always the case.

Usage
-----

Example::

   python csv2db.py -b sqlite < stdin 

Type::

   python csv2db.py --help

for command line help.

Command line options
--------------------

'''

import sys
from cgatcore.csv2db import main

import csv
csv.field_size_limit(sys.maxsize)


if __name__ == "__main__":
    sys.exit(main())



'''
csv_set.py - set operations on a table
======================================

:Tags: Python

Purpose
-------

.. todo::
   
   describe purpose of the script.

Usage
-----

Example::

   python csv_set.py --help

Type::

   python csv_set.py --help

for command line help.

Command line options
--------------------

'''
import sys

import cgatcore.experiment as E
from cgatcore.csvutils import readTable
import hashlib


class UniqueBuffer:
    mKeys = {}

    def __init__(self, outfile):
        self.mOutfile = outfile

    def write(self, out):
        key = hashlib.md5(out).digest()
        if key not in self.mKeys:
            self.mKeys[key] = True
            self.mOutfile.write(out)


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if argv is None:
        argv = sys.argv

    parser = E.OptionParser(
        version="%prog version: $Id: csv_set.py 2782 2009-09-10 11:40:29Z andreas $")

    parser.add_option("-u", "--unique", dest="unique", action="store_true",
                      help="output rows are uniq.")

    parser.add_option("-1", "--join-fields1", dest="join_fields1", type="string",
                      help="join fields in first table.")
    parser.add_option("-2", "--join-fields2", dest="join_fields2", type="string",
                      help="join fields in second table.")
    parser.add_option("-m", "--method", dest="method", type="choice",
                      help="set operation to perform.", choices=("intersection", "rest", "union"))

    parser.set_defaults(
        remove=False,
        unique=False,
        join_fields1=None,
        join_fields2=None,
        method="intersection",
    )

    (options, args) = E.start(parser, add_csv_options=True)

    if len(args) != 2:
        raise ValueError("please specify two files to join")

    if not options.join_fields1 or not options.join_fields2:
        raise ValueError("please specify at least one join field per table")

    options.join_fields1 = options.join_fields1.split(",")
    options.join_fields2 = options.join_fields2.split(",")

    options.filename1, options.filename2 = args

    fields1, table1 = readTable(open(options.filename1, "r"))
    fields2, table2 = readTable(open(options.filename2, "r"))

    if options.unique:
        outfile = UniqueBuffer(sys.stdout)
    else:
        outfile = options.stdout

    nfields1 = []
    for x in range(len(fields1)):
        if fields1[x] in options.join_fields1:
            nfields1.append(x)
    nfields2 = []
    for x in range(len(fields2)):
        if fields2[x] in options.join_fields2:
            nfields2.append(x)

    # calculate row indices: double keys are not taken care of here
    keys = {}
    for row1 in table1:
        v = [row1[x] for x in nfields1]
        key = hashlib.md5("".join(v)).digest()
        keys[key] = row1

    if options.method == "intersection":
        # build new field list
        take = list(range(len(fields1)))
        c = len(take)
        for x in fields2:
            if x not in options.join_fields2:
                take.append(c)
            c += 1

        t = fields1 + fields2

        new_fields = [t[x] for x in take]

        print("\t".join(new_fields))

        for row2 in table2:
            v = [row2[x] for x in nfields2]
            key = hashlib.md5("".join(v)).digest()
            if key in keys:
                new_row = keys[key] + row2
                outfile.write(
                    "\t".join([new_row[x] for x in take]) + "\n")

    elif options.method == "rest":

        new_fields = fields2
        print("\t".join(new_fields))

        for row2 in table2:
            v = [row2[x] for x in nfields2]
            key = hashlib.md5("".join(v)).digest()
            if key not in keys:
                outfile.write("\t".join(row2) + "\n")

    E.stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''
split_fasta.py - 
======================================================

:Tags: Python

Purpose
-------

.. todo::
   
   describe purpose of the script.

Usage
-----

Example::

   python split_fasta.py --help

Type::

   python split_fasta.py --help

for command line help.

Command line options
--------------------

'''
import sys
import re
import os
import cgat.FastaIterator as FastaIterator
import cgatcore.iotools as iotools
import cgatcore.experiment as E


class Files:

    mFiles = {}

    def __init__(self,
                 output_pattern=None,
                 skip_identifiers=False):

        self.mOutputPattern = output_pattern
        self.mSkipIdentifiers = skip_identifiers
        self.mCounts = {}

    def __del__(self):
        """close all open files."""
        for file in list(self.mFiles.values()):
            file.close()

    def GetFile(self, identifier):
        return identifier

    def GetFilename(self, identifier):
        """get filename for an identifier."""

        if self.mOutputPattern:
            return re.sub("%s", str(identifier), self.mOutputPattern)
        else:
            return identifier

    def OpenFile(self, filename, mode="w"):
        """open file.

        If file is in a new directory, create directories.
        """
        if mode in ("w", "a"):
            dirname = os.path.dirname(filename)
            if dirname and not os.path.exists(dirname):
                os.makedirs(dirname)

        returniotools.open_file(filename, mode)

    def Write(self, identifier, sequence):

        filename = self.GetFilename(identifier)

        if filename not in self.mFiles:

            if len(self.mFiles) > 1000:
                for f in list(self.mFiles.values()):
                    f.close()
                self.mFiles = {}

            self.mFiles[filename] = self.OpenFile(filename, "a")

        if self.mSkipIdentifiers:
            self.mFiles[filename].write("%s\n" % (sequence.sequence))
        else:
            self.mFiles[filename].write(
                ">%s\n%s\n" % (sequence.title, sequence.sequence))

        if filename not in self.mCounts:
            self.mCounts[filename] = 0
        self.mCounts[filename] += 1

    def DeleteFiles(self, min_size=0):
        """delete all files below a minimum size."""

        ndeleted = 0
        for filename, counts in list(self.mCounts.items()):
            if counts < min_size:
                os.remove(filename)
                ndeleted += 1

        return ndeleted


class FilesChunks(Files):

    def __init__(self,
                 chunk_size, **kwargs):

        Files.__init__(self, **kwargs)
        self.mChunkSize = chunk_size
        self.mFilename = 0

    def GetFilename(self, identifier):

        if not self.mFilename or self.mCounts[self.mFilename] % self.mChunkSize == 0:
            self.mFilename = re.sub(
                "%s", str(len(self.mCounts) + 1), self.mOutputPattern)

        return self.mFilename


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if argv is None:
        argv = sys.argv

    parser = E.OptionParser(
        version="%prog version: $Id: split_fasta.py 1714 2007-12-11 16:51:12Z andreas $")

    parser.add_option("-f", "--file", dest="input_filename", type="string",
                      help="input filename. If not given, stdin is used.",
                      metavar="FILE")

    parser.add_option("-i", "--input-pattern", dest="input_pattern", type="string",
                      help="input pattern. Parses description line in order to extract id.")

    parser.add_option("-o", "--output-filename-pattern", dest="output_pattern", type="string",
                      help="output pattern. Gives filename for a given sequence.")

    parser.add_option("-n", "--num-sequences", dest="num_sequences", type="int",
                      help="split by number of sequences (not implemented yet).")

    parser.add_option("-m", "--map", dest="map_filename", type="string",
                      help="map filename. Map identifiers to filenames",
                      metavar="FILE")

    parser.add_option("-s", "--skip-identifiers", dest="skip_identifiers", action="store_true",
                      help="do not write identifiers.",
                      metavar="FILE")

    parser.add_option("--min-size", dest="min_size", type="int",
                      help="minimum cluster size.")

    parser.set_defaults(
        input_filename=None,
        map_filename=None,
        skip_identifiers=False,
        input_pattern="^(\S+)",
        min_size=0,
        num_sequences=None,
        output_pattern="%s")

    (options, args) = E.start(parser)

    if options.input_filename:
        infile = iotools.open_file(options.input_filename, "r")
    else:
        infile = sys.stdin

    if options.map_filename:
        map_id2filename = iotools.ReadMap(open(options.map_filename, "r"))
    else:
        map_id2filename = {}

    if options.num_sequences:
        files = FilesChunks(chunk_size=options.num_sequences,
                            output_pattern=options.output_pattern,
                            skip_identifiers=options.skip_identifiers)

    else:
        files = Files(output_pattern=options.output_pattern,
                      skip_identifiers=options.skip_identifiers)

    if options.input_pattern:
        rx = re.compile(options.input_pattern)
    else:
        rx = None

    ninput = 0
    noutput = 0
    identifier = None
    chunk = 0

    for seq in FastaIterator.iterate(infile):

        ninput += 1

        if rx:
            try:
                identifier = rx.search(seq.title).groups()[0]
            except AttributeError:
                print("# parsing error in description line %s" % (seq.title))
        else:
            identifier = seq.title

        if map_id2filename:
            if identifier in map_id2filename:
                identifier = map_id2filename[identifier]
            else:
                continue

        files.Write(identifier, seq)
        noutput += 1

    if options.input_filename:
        infile.close()

    # delete all clusters below a minimum size
    # Note: this has to be done at the end, because
    # clusters sizes are only available once both the fasta
    # file and the map has been parsed.
    if options.min_size:
        ndeleted = files.DeleteFiles(min_size=options.min_size)
    else:
        ndeleted = 0

    if options.loglevel >= 1:
        print("# input=%i, output=%i, ndeleted=%i" % (ninput, noutput, ndeleted))

    E.stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""

This script currently only processes pairwise MAF alignments.

"""

import sys
import re
import collections
import cgatcore.experiment as E
import cgatcore.iotools as iotools
from cgat.Genomics import parse_region_string


def iterate_maf_blocks(infile):
    block = []
    for line in infile:
        if line.startswith("a"):
            if block:
                yield block
            block = []
        block.append(line)
    yield block

# The following is derived from bx.python (MIT licence)


def format_tabular(rows, align=None):
    if len(rows) == 0:
        return ""
    lengths = [len(col) for col in rows[0]]
    for row in rows[1:]:
        for i in range(0, len(row)):
            lengths[i] = max(lengths[i], len(row[i]))
    rval = ""
    for row in rows:
        for i in range(0, len(row)):
            if align and align[i] == "l":
                rval += row[i].ljust(lengths[i])
            else:
                rval += row[i].rjust(lengths[i])
            rval += " "
        rval += "\n"
    return rval


def parse_block(block):
    RECORD = collections.namedtuple(
        "RECORD",
        ("key", "src", "start", "size", "strand", "srcsize", "text"))

    key, src, start, size, strand, srcsize, text = re.split("\s+", block[1].strip())
    query = RECORD(key, src, int(start), int(size), strand, int(srcsize), text)
    key, src, start, size, strand, srcsize, text = re.split("\s+", block[2].strip())
    target = RECORD(key, src, int(start), int(size), strand, int(srcsize), text)
    try:
        key, text = re.split("\s+", block[3].strip())
    except ValueError:
        qual = None
    return block[0], query, target, qual


def main(argv=sys.argv):

    parser = E.OptionParser(version="%prog version: $Id$",
                            usage=globals()["__doc__"])

    parser.add_option(
        "-f", "--input-filter-tsv",
        dest="input_filter_tsv", type="string",
        help="list with identifiers to remove. "
        "[%default]")

    parser.add_option(
        "--set-prefix", dest="set_prefix", type="string",
        help="set sequence prefix [%default]")

    parser.add_option(
        "--min-length", dest="min_length", type="int",
        help="minimum alignment length [%default]")

    parser.add_option(
        "--method", dest="methods", action="append",
        choices=("shift-region", ),
        help="methods to apply [%default]")

    parser.set_defaults(
        input_maf_file=None,
        input_filter_tsv=None,
        set_prefix=None,
        min_length=0,
        methods=[],
    )

    (options, args) = E.start(parser, argv)

    if options.input_filter_tsv:
        with iotools.open_file(options.input_filter_tsv) as inf:
            skip_id = set([x[:-1] for x in inf])
    else:
        skip_id = False

    counter = E.Counter()

    if options.set_prefix:
        prefix = "s {}".format(options.set_prefix)
    else:
        prefix = None

    for block in iterate_maf_blocks(options.stdin):
        counter.blocks_input += 1
        if skip_id:
            if block[2].startswith("s "):
                id = re.match("s (\S+)", block[2]).groups()[0]
                if id in skip_id:
                    counter.blocks_skipped_id += 1
                    continue

        if options.min_length:
            if block[2].startswith("s "):
                id, pos, length = re.match("s (\S+)\s+(\d+)\s+(\d+)", block[2]).groups()
                if int(length) <= options.min_length:
                    counter.blocks_skipped_length += 1
                    continue

        if prefix:
            block[2] = prefix + block[2][4:]

        if block[2].startswith("s "):
            header, ali1, ali2, qual = parse_block(block)
            if "shift-region" in options.methods:
                rows = []
                contig, start, end = parse_region_string(ali1.src)
                ali1 = ali1._replace(src=contig, start=start + ali1.start)
                rows.append(list(map(str, ali1)))
                rows.append(list(map(str, ali2)))
                if qual:
                    rows.append(list(map(str, qual)))
                lines = [header]
                lines.append(format_tabular(rows, "llrrrrl"))
                lines.append("\n")
                block = lines
        counter.blocks_output += 1
        options.stdout.write("".join(block))

    E.info(counter)
    E.stop()



'''
csv_select.py - select rows from a table
========================================

:Tags: Python

Purpose
-------

extract rows from a csv-formatted table.

The select statement is a one-line, for example::

   csv_select.py "int(r['mC-foetal-sal-R4']) > 0" < in > out

Note the required variable name r for denoting field names. Please
also be aware than numeric values need to be converted first before
testing.

Usage
-----

Type::

   python csv_select.py --help

for command line help.

Command line options
--------------------

'''
import sys
import csv
import _csv
import cgatcore.experiment as E
from cgatcore.csvutils import CommentStripper, DictReaderLarge


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if argv is None:
        argv = sys.argv

    parser = E.OptionParser(version="%prog version: $Id: csv_cut.py 2782 2009-09-10 11:40:29Z andreas $",
                            usage=globals()["__doc__"])

    parser.add_option("-r", "--remove", dest="remove", action="store_true",
                      help="remove specified columns, keep all others.")

    parser.add_option("-u", "--unique", dest="unique", action="store_true",
                      help="output rows are uniq.")

    parser.add_option("-l", "--large", dest="large", action="store_true",
                      help="large columns. Do not use native python csv module [default=%default].")

    parser.add_option("-f", "--filename-fields", dest="filename_fields", type="string",
                      help="filename with field information.")

    parser.set_defaults(
        remove=False,
        unique=False,
        filename_fields=None,
    )

    (options, args) = E.start(parser,
                              add_csv_options=True,
                              quiet=True)

    statement = " ".join(args)

    if options.large:
        reader = DictReaderLarge(CommentStripper(sys.stdin),
                                 dialect=options.csv_dialect)
    else:
        reader = csv.DictReader(CommentStripper(sys.stdin),
                                dialect=options.csv_dialect)

    exec("f = lambda r: %s" % statement, globals())
    counter = E.Counter()
    writer = csv.DictWriter(options.stdout,
                            reader.fieldnames,
                            dialect=options.csv_dialect,
                            lineterminator=options.csv_lineterminator)

    writer.writerow(dict((fn, fn) for fn in reader.fieldnames))
    while 1:
        counter.input += 1
        try:
            row = next(reader)
        except _csv.Error as msg:
            options.stderr.write("# error while parsing: %s\n" % (msg))
            counter.errors += 1
            continue
        except StopIteration:
            break

        if not row:
            break

        if f(row):
            writer.writerow(row)
            counter.output += 1
        else:
            counter.filtered += 1

    E.info("%s" % counter)

    E.stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))
